import unittest

import elegy
import jax
import jax.numpy as jnp
from elegy.generalized_module.generalized_module import generalize
import haiku
import numpy as np


class TestHaikuModule(unittest.TestCase):
    def test_basic(self):
        class M(haiku.Module):
            def __call__(self, x):

                n = haiku.get_state(
                    "n", shape=[], dtype=jnp.int32, init=lambda *args: np.array(0)
                )
                w = haiku.get_parameter("w", [], init=lambda *args: np.array(2.0))

                haiku.set_state("n", n + 1)

                return x * w

        def f(x, initializing, rng):
            return M()(x)

        gm = elegy.HaikuModule(f)
        rng = elegy.RNGSeq(42)

        y_true, params, states = gm.init(rng)(x=3.0, y=1, rng=None, initializing=True)

        assert y_true == 6
        assert params["m"]["w"] == 2
        assert states["m"]["n"] == 0

        params = haiku.data_structures.to_mutable_dict(params)
        params["m"]["w"] = np.array(10.0)
        y_true, params, states = gm.apply(params, states, training=True, rng=rng)(
            x=3.0, y=1, rng=None, initializing=True
        )

        assert y_true == 30
        assert params["m"]["w"] == 10
        assert states["m"]["n"] == 1

    def test_summaries(self):
        class ModuleC(haiku.Module):
            @elegy.haiku_summarize
            def __call__(self, x):
                c1 = haiku.get_parameter("c1", [5], jnp.int32, init=jnp.ones)
                c2 = haiku.get_state("c2", [6], jnp.int32, init=jnp.ones)

                x = jax.nn.relu(x)
                elegy.haiku_summary("relu", jax.nn.relu, x)

                return x

        class ModuleB(haiku.Module):
            @elegy.haiku_summarize
            def __call__(self, x):
                b1 = haiku.get_parameter("b1", [3], jnp.int32, init=jnp.ones)
                b2 = haiku.get_state("b2", [4], jnp.int32, init=jnp.ones)

                x = ModuleC()(x)

                x = jax.nn.relu(x)
                elegy.haiku_summary("relu", jax.nn.relu, x)

                return x

        class ModuleA(haiku.Module):
            @elegy.haiku_summarize
            def __call__(self, x):
                a1 = haiku.get_parameter("a1", [1], jnp.int32, init=jnp.ones)
                a2 = haiku.get_state("a2", [2], jnp.int32, init=jnp.ones)

                x = ModuleB()(x)

                x = jax.nn.relu(x)
                elegy.haiku_summary("relu", jax.nn.relu, x)

                return x

        def f(x):
            return ModuleA()(x)

        model = elegy.Model(elegy.HaikuModule(f))

        summary_text = model.summary(x=jnp.ones([10, 2]), depth=2, return_repr=True)
        assert summary_text is not None

        lines = summary_text.split("\n")

        assert "module_a/module_b" in lines[7]
        assert "ModuleB" in lines[7]
        assert "(10, 2)" in lines[7]
        assert "8" in lines[7]
        assert "32 B" in lines[7]
        assert "10" in lines[7]
        assert "40 B" in lines[7]

        assert "module_a/relu" in lines[9]
        assert "(10, 2)" in lines[9]

        assert "module_a" in lines[11]
        assert "ModuleA" in lines[11]
        assert "(10, 2)" in lines[11]
        assert "1" in lines[11]
        assert "4 B" in lines[11]
        assert "2" in lines[11]
        assert "8 B" in lines[11]

        assert "21" in lines[13]
        assert "84 B" in lines[13]

        assert "9" in lines[14]
        assert "36 B" in lines[14]

        assert "12" in lines[15]
        assert "48 B" in lines[15]



import typing as tp

from elegy.losses.loss import Reduction
from elegy.regularizers.global_l1l2 import GlobalL1L2


def GlobalL1(
    l: float = 0.01,
    reduction: tp.Optional[Reduction] = None,
    name: str = "l1_regularization",
    **kwargs
) -> GlobalL1L2:
    r"""
    Create a regularizer that applies an L1 regularization penalty.

    The L1 regularization penalty is computed as:

    $$\ell_1\,\,penalty =\ell_1\sum_{i=0}^n|x_i|$$

    Usage:

    ```python
    model = elegy.Model(
        module_fn,
        loss=[
            elegy.losses.SparseCategoricalCrossentropy(),
            elegy.regularizers.GlobalL1(l=1e-5)
        ],
        metrics=lambda: elegy.metrics.SparseCategoricalAccuracy(),
    )
    ```

    Arguments:
        l: L1 regularization factor.
        kwargs: Additional keyword arguments passed to Module.

    Returns:
        An L1 Regularizer with the given regularization factor.
    """
    return GlobalL1L2(l1=l, reduction=reduction, name=name, **kwargs)



import typing as tp

import jax.numpy as jnp
import numpy as np

from elegy import types, utils
from elegy.metrics.accuracy import accuracy
from elegy.metrics.mean import Mean


def binary_accuracy(
    y_true: np.ndarray, y_pred: np.ndarray, threshold: float = 0.5
) -> np.ndarray:
    """
    Calculates how often predictions matches binary labels.

    Standalone usage:

    ```python
    y_true = np.array([[1], [1], [0], [0]])
    y_pred = np.array([[1], [1], [0], [0]])
    m = elegy.metrics.binary_accuracy(y_true, y_pred)
    assert m.shape == (4,)
    ```
    Arguments:
        y_true: Ground truth values. shape = `[batch_size, d0, .. dN]`.
        y_pred: The predicted values. shape = `[batch_size, d0, .. dN]`.
        threshold: Float representing the threshold for deciding whether
            prediction values are 1 or 0.

    Returns:
        Binary accuracy values. shape = `[batch_size, d0, .. dN-1]`
    """
    assert abs(y_pred.ndim - y_true.ndim) <= 1

    y_true, y_pred = utils.maybe_expand_dims(y_true, y_pred)

    y_pred = y_pred > threshold
    return jnp.mean(y_true == y_pred, axis=-1)


class BinaryAccuracy(Mean):
    """
    Calculates how often predictions matches binary labels.

    Standalone usage:
    ```python
    m = elegy.metrics.BinaryAccuracy()
    result = m(
        y_true=np.array([[1], [1], [0], [0]]),
        y_pred=np.array([[0.98], [1], [0], [0.6]]),
    )
    assert result == 0.75

    m = elegy.metrics.BinaryAccuracy()
    result = m(
        y_true=np.array([[1], [1], [0], [0]]),
        y_pred=np.array([[0.98], [1], [0], [0.6]]),
        sample_weight=np.array([1, 0, 0, 1]),
    )
    assert result == 0.5
    ```
    Usage with `Model` API:
    ```python
    model = elegy.Model(
        ...
        metrics=[tf.keras.metrics.BinaryAccuracy()],
    )
    ```
    """

    def __init__(
        self, threshold: float = 0.5, on: tp.Optional[types.IndexLike] = None, **kwargs
    ):
        """
        Creates a `BinaryAccuracy` instance.

        Arguments:
            threshold: Float representing the threshold for deciding whether
                prediction values are 1 or 0.
            on: A string or integer, or iterable of string or integers, that
                indicate how to index/filter the `y_true` and `y_pred`
                arguments before passing them to `call`. For example if `on = "a"` then
                `y_true = y_true["a"]`. If `on` is an iterable
                the structures will be indexed iteratively, for example if `on = ["a", 0, "b"]`
                then `y_true = y_true["a"][0]["b"]`, same for `y_pred`. For more information
                check out [Keras-like behavior](https://poets-ai.github.io/elegy/guides/modules-losses-metrics/#keras-like-behavior).
            kwargs: Additional keyword arguments passed to Module.
        """
        super().__init__(on=on, **kwargs)
        self.threshold = threshold

    def call(
        self,
        y_true: jnp.ndarray,
        y_pred: jnp.ndarray,
        sample_weight: tp.Optional[jnp.ndarray] = None,
    ) -> jnp.ndarray:
        """
        Accumulates metric statistics. `y_true` and `y_pred` should have the same shape.

        Arguments:
            y_true: Ground truth values. shape = `[batch_size, d0, .. dN]`.
            y_pred: The predicted values. shape = `[batch_size, d0, .. dN]`.
            sample_weight: Optional `sample_weight` acts as a
                coefficient for the metric. If a scalar is provided, then the metric is
                simply scaled by the given value. If `sample_weight` is a tensor of size
                `[batch_size]`, then the metric for each sample of the batch is rescaled
                by the corresponding element in the `sample_weight` vector. If the shape
                of `sample_weight` is `[batch_size, d0, .. dN-1]` (or can be broadcasted
                to this shape), then each metric element of `y_pred` is scaled by the
                corresponding value of `sample_weight`. (Note on `dN-1`: all metric
                functions reduce by 1 dimension, usually the last axis (-1)).
        Returns:
            Array with the cumulative accuracy.
        """

        return super().call(
            values=binary_accuracy(
                y_true=y_true, y_pred=y_pred, threshold=self.threshold
            ),
            sample_weight=sample_weight,
        )



import math
from unittest import TestCase

import jax.numpy as jnp
import numpy as np
import tensorflow as tf
from elegy.data.tf_dataset_adapter import TFDatasetAdapter


class ArrayDataAdapterTest(TestCase):
    def test_basic(self):
        batch_size = 10
        epochs = 1
        x = np.array(np.random.uniform(size=(100, 32, 32, 3)))
        y = np.array(np.random.uniform(size=(100, 1)))
        dataset = tf.data.Dataset.from_tensor_slices((x, y))
        dataset = dataset.batch(batch_size)

        data_adapter = TFDatasetAdapter(dataset, steps=None)

        num_steps = math.ceil(x.shape[0] / batch_size) * epochs
        iterator_fn = data_adapter.get_dataset()
        for i, batch in zip(range(num_steps), iterator_fn()):
            batch_x, batch_y = batch
            assert batch_x.shape == (batch_size, *x.shape[1:])
            assert batch_y.shape == (batch_size, *y.shape[1:])
            np.testing.assert_array_equal(
                batch_x, x[i * batch_size : (i + 1) * batch_size]
            )

        assert data_adapter.get_size() * batch_size == x.shape[0]
        assert data_adapter.batch_size == batch_size

    def test_only_x_repeat(self):
        batch_size = 10
        epochs = 2

        x = np.array(np.random.uniform(size=(100, 32, 32, 3)))
        dataset = tf.data.Dataset.from_tensor_slices(x)
        dataset = dataset.batch(batch_size)
        dataset = dataset.repeat()

        dataset_length = x.shape[0]
        num_steps = math.ceil(dataset_length / batch_size) * epochs

        data_adapter = TFDatasetAdapter(
            dataset, steps=math.ceil(dataset_length / batch_size)
        )

        iterator_fn = data_adapter.get_dataset()
        for i, batch in zip(range(num_steps), iterator_fn()):
            batch_x = batch
            assert batch_x.shape == (batch_size, *x.shape[1:])
            np.testing.assert_array_equal(
                batch_x,
                x[
                    (i * batch_size)
                    % dataset_length : (i * batch_size)
                    % dataset_length
                    + batch_size
                ],
            )

        assert data_adapter.get_size() * batch_size == x.shape[0]
        assert data_adapter.batch_size == batch_size
        assert i == num_steps - 1

    def test_error(self):
        batch_size = 10
        epochs = 2
        x = np.array(np.random.uniform(size=(100, 32, 32, 3)))
        dataset = tf.data.Dataset.from_tensor_slices(x)
        dataset = dataset.batch(batch_size)

        data_adapter = TFDatasetAdapter(dataset, steps=None)

        num_steps = math.ceil(x.shape[0] / batch_size) * epochs
        iterator_fn = data_adapter.get_dataset()
        iterator = iterator_fn()

        with self.assertRaises(StopIteration):
            for i in range(num_steps):
                batch = next(iterator)
                batch_x = batch
                assert batch_x.shape == (batch_size, *x.shape[1:])
                np.testing.assert_array_equal(
                    batch_x, x[i * batch_size : (i + 1) * batch_size]
                )



from jax._src.lax.lax import remaining
from torch.utils.data import DataLoader

from .data_adapter import DataAdapter
from .utils import is_none_or_empty, map_structure, list_to_tuple


class TorchDataLoaderAdapter(DataAdapter):
    """Adapter that handles torch Dataloaders."""

    @staticmethod
    def can_handle(x, y=None):
        return isinstance(x, DataLoader)

    def __init__(
        self,
        x: DataLoader,
        y=None,
        steps=None,
        sample_weights=None,
        training=True,
        **kwargs,
    ):

        if not is_none_or_empty(y):
            raise ValueError(
                "`y` argument is not supported when using " "torch Dataloader as input."
            )
        if not is_none_or_empty(sample_weights):
            raise ValueError(
                "`sample_weight` argument is not supported when using "
                "torch Dataloader as input."
            )

        super().__init__(x, y, **kwargs)

        self.training = training
        self.steps = steps
        self._batch_size = x.batch_size
        self._dataset = x

        self.current_step = 0

    def get_dataset(self):
        def parse_dataloader_gen():
            self.current_step = 0
            for batch in iter(self._dataset):
                self.current_step += 1
                batch = map_structure(lambda x: x.cpu().numpy(), list_to_tuple(batch))
                yield batch

        return parse_dataloader_gen

    def get_size(self):
        try:
            return len(self._dataset)
        except Exception:
            return None

    @property
    def batch_size(self):
        return self.representative_batch_size

    @property
    def representative_batch_size(self):
        return self._batch_size

    def has_partial_batch(self):
        return False

    @property
    def partial_batch_size(self):
        return

    def should_recreate_iterator(self):
        # if in eval mode should not recreate iterator
        # but if in train mode and steps not provided, should recreate at end of each epoch
        if not self.training or self.steps is None:
            return self.training

        steps_dataset = self.get_size()
        if steps_dataset is None:
            return False

        remaining_steps = steps_dataset - self.current_step
        # if remaining steps less than needed steps, should recreate dataloader
        # TODO: This will drop the last steps of data, how to avoid this?
        if remaining_steps < self.steps:
            return True
        else:
            return False



# Implementation based on tf.keras.engine.data_adapter.py
# https://github.com/tensorflow/tensorflow/blob/2b96f3662bd776e277f86997659e61046b56c315/tensorflow/python/keras/engine/data_adapter.py


import collections
import functools
import math
import typing as tp

import jax.numpy as jnp
import numpy as np
import six
from elegy import types, utils


class Multimap(types.Protocol):
    def __call__(self, *args: types.np.ndarray) -> types.T:
        ...


def map_append(output, batch_output):
    output.append(batch_output)
    return output


def map_structure(
    f: Multimap,
    *structure: tp.Union[types.ArrayHolder, None],
) -> types.Container[tp.Union[types.T, None]]:

    if isinstance(structure[0], tp.Tuple):
        return tuple(map_structure(f, *x) for x in zip(*structure))
    elif isinstance(structure[0], tp.Dict):
        return {
            key: map_structure(f, *(elem[key] for elem in structure))
            for key in structure[0]
        }
    elif structure[0] is not None:
        return f(*structure)
    else:  # if it is tuple of None
        return None


def flatten(inputs: types.ArrayHolder) -> tp.Iterable[types.np.ndarray]:

    if isinstance(inputs, (jnp.ndarray, np.ndarray)):
        yield inputs
    elif isinstance(inputs, tp.Tuple):
        for x in inputs:
            yield from flatten(x)
    elif isinstance(inputs, tp.Dict):
        for x in inputs.values():
            yield from flatten(x)
    elif isinstance(inputs, (tp.Generator, tp.Iterator, tp.Iterable)):
        yield inputs
    else:
        raise TypeError(f"Unsupported type '{type(inputs)}'")


def pack_x_y_sample_weight(x, y=None, sample_weight=None):
    """Packs user-provided data into a tuple."""
    if y is None:
        return (x,)
    elif sample_weight is None:
        return (x, y)
    else:
        return (x, y, sample_weight)


def unpack_x_y_sample_weight(data):
    """Unpacks user-provided data tuple."""
    if not isinstance(data, tuple):
        return (data, None, None)
    elif len(data) == 1:
        return (data[0], None, None)
    elif len(data) == 2:
        return (data[0], data[1], None)
    elif len(data) == 3:
        return (data[0], data[1], data[2])

    raise ValueError("Data not understood.")


def list_to_tuple(maybe_list):
    if isinstance(maybe_list, list):
        return tuple(maybe_list)
    return maybe_list


def handle_partial_sample_weights(y, sample_weights):
    any_sample_weight = sample_weights is not None and any(
        w is not None for w in sample_weights
    )
    partial_sample_weight = any_sample_weight and any(w is None for w in sample_weights)

    if not any_sample_weight:
        return None

    if not partial_sample_weight:
        return sample_weights


def is_none_or_empty(inputs):
    # util method to check if the input is a None or a empty list.
    # the python "not" check will raise an error like below if the input is a
    # numpy array
    # "The truth value of an array with more than one element is ambiguous.
    # Use a.any() or a.all()"
    return inputs is None or not list(flatten(inputs))


def assert_not_namedtuple(x):
    if (
        isinstance(x, tuple)
        and
        # TODO(b/144192902): Use a namedtuple checking utility.
        hasattr(x, "_fields")
        and isinstance(x._fields, collections.Sequence)
        and all(isinstance(f, six.string_types) for f in x._fields)
    ):
        raise ValueError(
            "Received namedtuple ({}) with fields `{}` as input. namedtuples "
            "cannot, in general, be unambiguously resolved into `x`, `y`, "
            "and `sample_weight`. For this reason Keras has elected not to "
            "support them. If you would like the value to be unpacked, "
            "please explicitly convert it to a tuple before passing it to "
            "Keras.".format(x.__class__, x._fields)
        )


def train_validation_split(arrays, validation_split, shuffle=True) -> tp.Tuple:
    """
    Split arrays into random train and validation subsets.
    Arguments:
        arrays: Arrays to split. Allowed inputs are arbitrarily nested structures
            of Jax and NumPy arrays.
        validation_split: Float between 0 and 1. The proportion of the dataset to
            include in the validation split. The rest of the dataset will be included
            in the training split.
        shuffle: Bool. Whether to shuffle the data before performing a split. If
            `False`, the last `validation_split` fraction of that training data will
            become the validation split.
    Returns:
        `(train_arrays, validation_arrays)`
    """

    def _can_split(t):
        supported_types = (jnp.ndarray, np.ndarray)
        # if pd:
        #     supported_types = (jnp.ndarray, np.ndarray, pd.Series, pd.DataFrame)
        return isinstance(t, supported_types) or t is None

    flat_arrays = flatten(arrays)
    # flat_arrays = arrays
    if not all(_can_split(t) for t in arrays):
        raise ValueError(
            "`validation_split` is only supported for Tensors or NumPy "
            "arrays, found: {}".format(arrays)
        )

    if all(t is None for t in flat_arrays):
        return arrays, arrays

    first_non_none = None
    for t in flat_arrays:
        if t is not None:
            first_non_none = t
            break

    # Assumes all arrays have the same batch shape or are `None`.
    batch_dim = int(first_non_none.shape[0])
    indices = np.arange(batch_dim)
    if shuffle:
        indices = np.random.shuffle(indices)
    split_at = int(math.floor(batch_dim * (1.0 - validation_split)))
    train_indices = indices[:split_at]
    val_indices = indices[split_at:]

    def _split(t, indices):
        if t is None:
            return t
        return t[indices]

    train_arrays = map_structure(
        functools.partial(_split, indices=train_indices), arrays
    )
    val_arrays = map_structure(functools.partial(_split, indices=val_indices), arrays)
    return train_arrays, val_arrays



from elegy import types
import typing as tp
import jax.numpy as jnp
from elegy import utils, types
from elegy.losses.loss import Loss, Reduction


def mean_absolute_percentage_error(
    y_true: jnp.ndarray, y_pred: jnp.ndarray
) -> jnp.ndarray:
    """
    Computes the mean absolute percentage error (MAPE) between labels and predictions.

    After computing the absolute distance between the true value and the prediction value
    and divide by the true value, the mean value over the last dimension is returned.

    Usage:

    ```python
    rng = jax.random.PRNGKey(42)

    y_true = jax.random.randint(rng, shape=(2, 3), minval=0, maxval=2)
    y_pred = jax.random.uniform(rng, shape=(2, 3))

    loss = elegy.losses.mean_absolute_percentage_error(y_true, y_pred)

    assert loss.shape == (2,)

    assert jnp.array_equal(loss, 100. * jnp.mean(jnp.abs((y_pred - y_true) / jnp.clip(y_true, types.EPSILON, None))))
    ```

    Arguments:
        y_true: Ground truth values. shape = `[batch_size, d0, .. dN]`.
        y_pred: The predicted values. shape = `[batch_size, d0, .. dN]`.

    Returns:
        Mean absolute percentage error values. shape = `[batch_size, d0, .. dN-1]`.
    """
    y_true = y_true.astype(y_pred.dtype)
    diff = jnp.abs((y_pred - y_true) / jnp.maximum(jnp.abs(y_true), types.EPSILON))
    return 100.0 * jnp.mean(diff, axis=-1)


class MeanAbsolutePercentageError(Loss):
    """
    Computes the mean absolute errors between labels and predictions.

    `loss = mean(abs((y_true - y_pred) / y_true))`

    Usage:

    ```python
    y_true = jnp.array([[1.0, 1.0], [0.9, 0.0]])
    y_pred = jnp.array([[1.0, 1.0], [1.0, 0.0]])

    # Using 'auto'/'sum_over_batch_size' reduction type.
    mape = elegy.losses.MeanAbsolutePercentageError()
    result = mape(y_true, y_pred)
    assert jnp.isclose(result, 2.78, rtol=0.01)

    # Calling with 'sample_weight'.
    assert jnp.isclose(mape(y_true, y_pred, sample_weight=jnp.array([0.1, 0.9])), 2.5, rtol=0.01)

    # Using 'sum' reduction type.
    mape = elegy.losses.MeanAbsolutePercentageError(reduction=elegy.losses.Reduction.SUM)

    assert jnp.isclose(mape(y_true, y_pred), 5.6, rtol=0.01)

    # Using 'none' reduction type.
    mape = elegy.losses.MeanAbsolutePercentageError(reduction=elegy.losses.Reduction.NONE)

    assert jnp.all(jnp.isclose(result, [0. , 5.6], rtol=0.01))

    ```
    Usage with the Elegy API:

    ```python
    model = elegy.Model(
        module_fn,
        loss=elegy.losses.MeanAbsolutePercentageError(),
        metrics=elegy.metrics.Mean(),
    )
    ```
    """

    def __init__(
        self,
        reduction: tp.Optional[Reduction] = None,
        weight: tp.Optional[float] = None,
        on: tp.Optional[types.IndexLike] = None,
        **kwargs
    ):
        """
        Initializes `Mean` class.

        Arguments:
            reduction: (Optional) Type of `elegy.losses.Reduction` to apply to
                loss. Default value is `SUM_OVER_BATCH_SIZE`. For almost all cases
                this defaults to `SUM_OVER_BATCH_SIZE`. When used with
                `tf.distribute.Strategy`, outside of built-in training loops such as
                `elegy` `compile` and `fit`, or `SUM_OVER_BATCH_SIZE`
                will raise an error.
                for more details.
            weight: Optional weight contribution for the total loss. Defaults to `1`.
            on: A string or integer, or iterable of string or integers, that
                indicate how to index/filter the `y_true` and `y_pred`
                arguments before passing them to `call`. For example if `on = "a"` then
                `y_true = y_true["a"]`. If `on` is an iterable
                the structures will be indexed iteratively, for example if `on = ["a", 0, "b"]`
                then `y_true = y_true["a"][0]["b"]`, same for `y_pred`. For more information
                check out [Keras-like behavior](https://poets-ai.github.io/elegy/guides/modules-losses-metrics/#keras-like-behavior).
        """

        return super().__init__(reduction=reduction, weight=weight, on=on, **kwargs)

    def call(
        self,
        y_true: jnp.ndarray,
        y_pred: jnp.ndarray,
        sample_weight: tp.Optional[
            jnp.ndarray
        ] = None,  # not used, __call__ handles it, left for documentation purposes.
    ) -> jnp.ndarray:
        """
        Invokes the `MeanAbsolutePercentageError` instance.

        Arguments:
            y_true: Ground truth values. shape = `[batch_size, d0, .. dN]`, except
                sparse loss functions such as sparse categorical crossentropy where
                shape = `[batch_size, d0, .. dN-1]`
            y_pred: The predicted values. shape = `[batch_size, d0, .. dN]`
            sample_weight: Optional `sample_weight` acts as a
                coefficient for the loss. If a scalar is provided, then the loss is
                simply scaled by the given value. If `sample_weight` is a tensor of size
                `[batch_size]`, then the total loss for each sample of the batch is
                rescaled by the corresponding element in the `sample_weight` vector. If
                the shape of `sample_weight` is `[batch_size, d0, .. dN-1]` (or can be
                broadcasted to this shape), then each loss element of `y_pred` is scaled
                by the corresponding value of `sample_weight`. (Note on`dN-1`: all loss
                functions reduce by 1 dimension, usually axis=-1.)

        Returns:
            Weighted loss float `Tensor`. If `reduction` is `NONE`, this has
                shape `[batch_size, d0, .. dN-1]`; otherwise, it is scalar. (Note `dN-1`
                because all loss functions reduce by 1 dimension, usually axis=-1.)

        Raises:
            ValueError: If the shape of `sample_weight` is invalid.
        """
        return mean_absolute_percentage_error(y_true, y_pred)



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Layer Norm."""

import collections
from elegy import types
from typing import Optional, Sequence, Union

import jax
import jax.numpy as jnp
import numpy as np
from haiku._src import utils as hk_utils

# If you are forking replace this with `import haiku as hk`.
from elegy import initializers, module


class LayerNormalization(module.Module):
    """LayerNorm module.

    See: https://arxiv.org/abs/1607.06450.
    """

    def __init__(
        self,
        axis: Union[int, Sequence[int], slice] = -1,
        create_scale: bool = True,
        create_offset: bool = True,
        eps: float = 1e-5,
        scale_init: Optional[types.Initializer] = None,
        offset_init: Optional[types.Initializer] = None,
        **kwargs
    ):
        """Constructs a LayerNorm module.

        Args:
            axis: Integer, list of integers, or slice indicating which axes to
                normalize over.
            create_scale: Bool, defines whether to create a trainable scale
                per channel applied after the normalization.
            create_offset: Bool, defines whether to create a trainable offset
                per channel applied after normalization and scaling.
            eps: Small epsilon to avoid division by zero variance. Defaults ``1e-5``,
                as in the paper and Sonnet.
            scale_init: Optional initializer for gain (aka scale). By default, one.
            offset_init: Optional initializer for bias (aka offset). By default, zero.
            kwargs: Additional keyword arguments passed to Module.
        """
        super().__init__(**kwargs)
        if not create_scale and scale_init is not None:
            raise ValueError("Cannot set `scale_init` if `create_scale=False`.")
        if not create_offset and offset_init is not None:
            raise ValueError("Cannot set `offset_init` if `create_offset=False`.")

        if isinstance(axis, slice):
            self.axis = axis
        elif isinstance(axis, int):
            self.axis = (axis,)
        elif isinstance(axis, collections.Iterable) and all(
            isinstance(ax, int) for ax in axis
        ):
            self.axis = tuple(axis)
        else:
            raise ValueError("`axis` should be an int, slice or iterable of ints.")

        self.eps = eps
        self.create_scale = create_scale
        self.create_offset = create_offset
        self.scale_init = scale_init or jnp.ones
        self.offset_init = offset_init or jnp.zeros

    def call(
        self,
        inputs: jnp.ndarray,
        scale: Optional[jnp.ndarray] = None,
        offset: Optional[jnp.ndarray] = None,
    ) -> jnp.ndarray:
        """Connects the layer norm.

        Args:
          inputs: An array, where the data format is ``[N, ..., C]``.
          scale: An array up to n-D. The shape of this tensor must be broadcastable
            to the shape of ``inputs``. This is the scale applied to the normalized
            inputs. This cannot be passed in if the module was constructed with
            ``create_scale=True``.
          offset: An array up to n-D. The shape of this tensor must be broadcastable
            to the shape of ``inputs``. This is the offset applied to the normalized
            inputs. This cannot be passed in if the module was constructed with
            ``create_offset=True``.

        Returns:
          The array, normalized.
        """
        if self.create_scale and scale is not None:
            raise ValueError("Cannot pass `scale` at call time if `create_scale=True`.")
        if self.create_offset and offset is not None:
            raise ValueError(
                "Cannot pass `offset` at call time if `create_offset=True`."
            )

        axis = self.axis
        if isinstance(axis, slice):
            axis = tuple(range(inputs.ndim)[axis])

        mean = jnp.mean(inputs, axis=axis, keepdims=True)
        variance = jnp.var(inputs, axis=axis, keepdims=True)

        param_shape = inputs.shape[-1:]
        if self.create_scale:
            scale = self.add_parameter(
                "scale", param_shape, jnp.float32, initializer=self.scale_init
            )
        elif scale is None:
            scale = np.array(1.0, dtype=inputs.dtype)

        if self.create_offset:
            offset = self.add_parameter(
                "offset", param_shape, jnp.float32, initializer=self.offset_init
            )
        elif offset is None:
            offset = np.array(0.0, dtype=inputs.dtype)

        scale = jnp.broadcast_to(scale, inputs.shape)
        offset = jnp.broadcast_to(offset, inputs.shape)
        mean = jnp.broadcast_to(mean, inputs.shape)

        inv = scale * jax.lax.rsqrt(variance + self.eps)
        return inv * (inputs - mean) + offset


class InstanceNormalization(LayerNormalization):
    """Normalizes inputs along the spatial dimensions.

    See `LayerNorm` for more details.
    """

    def __init__(
        self,
        create_scale: bool = True,
        create_offset: bool = True,
        eps: float = 1e-5,
        scale_init: Optional[types.Initializer] = None,
        offset_init: Optional[types.Initializer] = None,
        data_format: str = "channels_last",
        **kwargs
    ):
        """Constructs an `InstanceNormalization` module.

        This method creates a module which normalizes over the spatial dimensions.

        Args:
            create_scale: ``bool`` representing whether to create a trainable scale
                per channel applied after the normalization.
            create_offset: ``bool`` representing whether to create a trainable offset
                per channel applied after normalization and scaling.
            eps: Small epsilon to avoid division by zero variance. Defaults to
                ``1e-5``.
            scale_init: Optional initializer for the scale variable. Can only be set
                if ``create_scale=True``. By default scale is initialized to ``1``.
            offset_init: Optional initializer for the offset variable. Can only be set
                if ``create_offset=True``. By default offset is initialized to ``0``.
            data_format: The data format of the input. Can be either
                ``channels_first``, ``channels_last``, ``N...C`` or ``NC...``. By
                default it is ``channels_last``.
            kwargs: Additional keyword arguments passed to Module.
        """
        if hk_utils.get_channel_index(data_format) == 1:
            axis = slice(2, None)
        else:  # channel_index = -1
            axis = slice(1, -1)
        super().__init__(
            axis=axis,
            create_scale=create_scale,
            create_offset=create_offset,
            eps=eps,
            scale_init=scale_init,
            offset_init=offset_init,
            **kwargs
        )



import functools
import hashlib
import inspect
import os
import re
import shutil
import sys
import typing as tp
import urllib
from functools import total_ordering

import jax
import jax.numpy as jnp
import jax.tree_util
import numpy as np
import toolz

from elegy import types


def maybe_expand_dims(a: np.ndarray, b: np.ndarray) -> tp.Tuple[np.ndarray, np.ndarray]:
    assert np.prod(a.shape) == np.prod(b.shape)

    if a.ndim < b.ndim:
        a = a[..., None]

    if b.ndim < a.ndim:
        b = b[..., None]

    return a, b


def wraps(f, docs: bool = True):
    assignments = ("__annotations__",)

    if docs:
        assignments += ("__doc__",)

    return functools.wraps(f, assigned=assignments, updated=())


def get_signature_f_recursive(f: tp.Callable) -> tp.Callable:

    if hasattr(f, "_signature_f"):
        return get_signature_f_recursive(f._signature_f)
    else:
        return f


def inject_dependencies(
    f: tp.Callable,
    signature_f: tp.Optional[tp.Callable] = None,
    rename: tp.Optional[tp.Dict[str, str]] = None,
):
    if signature_f is None:
        signature_f = f

    signature_f = get_signature_f_recursive(signature_f)
    f_params = get_function_args(signature_f)

    @functools.wraps(signature_f)
    def wrapper(*args, **kwargs):
        n_args = len(args)
        arg_names = [arg.name for arg in f_params[:n_args]]
        kwarg_names = [arg.name for arg in f_params[n_args:]]

        if rename:
            for old, new in rename.items():
                if old in kwargs:
                    kwargs[new] = kwargs.pop(old)

        if not any(arg.kind == inspect.Parameter.VAR_KEYWORD for arg in f_params):
            # print(list(kwargs.keys()))
            # print(kwarg_names)
            kwargs = {
                arg: kwargs[arg]
                for arg in kwarg_names
                if arg not in arg_names and arg in kwargs
            }

        return f(*args, **kwargs)

    return wrapper


def get_function_args(f) -> tp.List[inspect.Parameter]:
    return list(inspect.signature(f).parameters.values())


def get_input_args(
    x: tp.Union[np.ndarray, jnp.ndarray, tp.Dict[str, tp.Any], tp.Tuple],
    *,
    states: types.States,
    initializing: bool,
    mode: types.Mode,
) -> tp.Tuple[tp.Tuple, tp.Dict[str, tp.Any]]:

    if isinstance(x, tp.Tuple):
        args = x
        kwargs = {}
    elif isinstance(x, tp.Dict):
        args = ()
        kwargs = x
    else:
        args = (x,)
        kwargs = {}

    apply_kwargs = dict(
        initializing=initializing,
        net_params=states.net_params,
        net_states=states.net_states,
        rng=states.rng,
        states=states,
        mode=mode,
        training=(mode == types.Mode.train),
    )
    apply_kwargs.update(kwargs)

    return args, apply_kwargs


def lower_snake_case(s: str) -> str:
    s = re.sub(r"(?<!^)(?=[A-Z])", "_", s).lower()
    parts = s.split("_")
    output_parts = []

    for i in range(len(parts)):
        if i == 0 or len(parts[i - 1]) > 1:
            output_parts.append(parts[i])
        else:
            output_parts[-1] += parts[i]

    return "_".join(output_parts)


def get_name(obj) -> str:
    if hasattr(obj, "name") and obj.name:
        return obj.name
    elif hasattr(obj, "__name__") and obj.__name__:
        return obj.__name__
    elif hasattr(obj, "__class__") and obj.__class__.__name__:
        return lower_snake_case(obj.__class__.__name__)
    else:
        raise ValueError(f"Could not get name for: {obj}")


def _leaf_paths(
    path: types.Path, inputs: tp.Any
) -> tp.Iterable[tp.Tuple[types.Path, tp.Any]]:

    if isinstance(inputs, (tp.Tuple, tp.List)):
        for i, value in enumerate(inputs):
            yield from _leaf_paths(path + (i,), value)
    elif isinstance(inputs, tp.Dict):
        for name, value in inputs.items():
            yield from _leaf_paths(path + (name,), value)
    else:
        yield (path, inputs)


def leaf_paths(inputs: tp.Any) -> tp.List[tp.Tuple[types.Path, tp.Any]]:
    return list(_leaf_paths((), inputs))


def _flatten_names(
    path: types.Path, inputs: tp.Any
) -> tp.Iterable[tp.Tuple[types.Path, tp.Any]]:

    if isinstance(inputs, (tp.Tuple, tp.List)):
        for i, value in enumerate(inputs):
            yield from _flatten_names(path, value)
    elif isinstance(inputs, tp.Dict):
        for name, value in inputs.items():
            yield from _flatten_names(path + (name,), value)
    else:
        yield (path, inputs)


def flatten_names(inputs: tp.Any) -> tp.List[tp.Tuple[str, tp.Any]]:
    return [
        ("/".join(map(str, path)), value) for path, value in _flatten_names((), inputs)
    ]


def get_unique_name(
    names: tp.Set[str],
    name: str,
):

    if name in names:
        i = 1
        while f"{name}_{i}" in names:
            i += 1

        name = f"{name}_{i}"

    names.add(name)
    return name


def merge_with_unique_names(
    a: tp.Dict[str, tp.Any],
    *rest: tp.Dict[str, tp.Any],
) -> tp.Dict[str, tp.Any]:

    a = a.copy()

    for b in rest:
        a = _merge_with_unique_names(a, b)

    return a


def _merge_with_unique_names(
    a: tp.Dict[str, tp.Any],
    b: tp.Dict[str, tp.Any],
) -> tp.Dict[str, tp.Any]:
    names = set()
    output = dict(a)

    for name, value in b.items():
        output[get_unique_name(names, name)] = value

    return output


def parameters_count(params: tp.Any) -> int:
    leaves = (jnp.asarray(x) for x in jax.tree_leaves(params))
    return sum(x.size for x in leaves)


def parameters_bytes(params: tp.Any) -> int:
    leaves = (jnp.asarray(x) for x in jax.tree_leaves(params))
    return sum(x.size * x.dtype.itemsize for x in leaves)


def download_file(url, cache="~/.elegy/downloads", sha256=None):
    if cache.startswith("~/"):
        cache = os.path.join(os.path.expanduser("~"), cache[2:])
    cachefilename = os.path.basename(url)
    cachefilename = cachefilename[: cachefilename.find("?")]
    cachefilename = os.path.join(cache, cachefilename)

    if not os.path.exists(cachefilename):
        print(f"Downloading {url}")
        filename, _ = urllib.request.urlretrieve(url)
        if sha256 is not None:
            filehash = hashlib.sha256(open(filename, "rb").read()).hexdigest()
            if sha256 != filehash:
                raise RuntimeError("Downloaded file has an incorrect hash")
        os.makedirs(os.path.dirname(cachefilename), exist_ok=True)
        shutil.move(filename, cachefilename)

    return cachefilename


def merge_params(a: tp.Any, b: tp.Any):

    if isinstance(a, dict) and isinstance(b, dict):
        return {
            key: a[key]
            if key not in b
            else b[key]
            if key not in a
            else merge_params(a[key], b[key])
            for key in set(a) | set(b)
        }
    elif isinstance(a, list) and isinstance(b, list):
        if len(a) != len(b):
            raise ValueError(
                f"Cannot merge two lists of different lengths:\na={a}\nb={b}"
            )
        return [merge_params(a, b) for a, b in zip(a, b)]
    elif isinstance(a, tuple) and isinstance(b, tuple):
        if len(a) != len(b):
            raise ValueError(
                f"Cannot merge two tuples of different lengths:\na={a}\nb={b}"
            )
        return tuple(merge_params(a, b) for a, b in zip(a, b))
    else:
        raise ValueError(f"Cannot merge structs:\na={a}\nb={b}")


def get_path_params(path: types.Path, params: tp.Any) -> tp.Any:
    for key in path:
        try:
            params = params[key]
        except BaseException as e:
            return None

    return params


def none_or_uninitialized(x: tp.Any) -> bool:
    return x is None or isinstance(x, types.Uninitialized)


# def merge_collections(collections: types.ParameterCollection) -> tp.Dict[str, tp.Any]:
#     output_parameters = {}

#     for collection, values in collections.items():
#         params = jax.tree_map(lambda x: types.Parameter(x, collection), values)
#         output_parameters = merge_params(params, output_parameters)

#     assert isinstance(output_parameters, dict)

#     return output_parameters


def get_parameter(collections: types.ParameterCollection, name: str) -> types.Parameter:

    parameters = [
        (collection, parameters[name])
        for collection, parameters in collections.items()
        if name in parameters
    ]

    if len(parameters) == 0:
        raise ValueError(f"No parameters named {name} in collections {collections}")
    elif len(parameters) >= 2:
        raise ValueError(
            f"Multiple parameters named {name} in collections {collections}"
        )

    [(collection, value)] = parameters

    return types.Parameter(collection=collection, value=value)


def get_submodule_colletions(
    collections: types.ParameterCollection, name: str
) -> types.ParameterCollection:
    return {
        collection: parameters[name]
        for collection, parameters in collections.items()
        if name in parameters
    }


def split_into_collections(
    parameters: tp.Dict[str, tp.Any]
) -> types.ParameterCollection:
    all_collections = set()

    def find_collections(parameter: types.Parameter):
        all_collections.add(parameter.collection)

    jax.tree_map(find_collections, parameters)

    return {
        collection: unwrap_filter(
            lambda p: p.collection == collection,
            parameters,
        )
        for collection in all_collections
    }


def unwrap_filter(
    f: tp.Callable[[types.Parameter], bool], parameters: tp.Dict[str, tp.Any]
) -> tp.Dict[str, tp.Any]:
    outputs = {}

    for name, parameter in parameters.items():

        if isinstance(parameter, types.Parameter):
            if f(parameter):
                outputs[name] = parameter.value
        else:
            outputs[name] = unwrap_filter(f, parameter)

    return outputs


def plot_history(history):
    import matplotlib.pyplot as plt

    keys = [key for key in history.history.keys() if not key.startswith("val_")]
    n_plots = len(keys)

    figure = plt.figure(figsize=(14, 24))

    # for i, key in enumerate(list(history.history.keys())[:n_plots]):
    for i, key in enumerate(keys):
        if key == "size":
            continue

        metric = history.history[key]

        plt.subplot(n_plots, 1, i + 1)
        plt.plot(metric, label=f"Training {key}")

        try:
            val_metric = history.history[f"val_{key}"]
            plt.plot(val_metric, label=f"Validation {key}")
            title = f"Training and Validation {key}"
        except KeyError:
            title = f"Training {key}"

        plt.legend(loc="lower right")
        plt.ylabel(key)
        plt.title(title)

    plt.show()



import elegy
import jax.numpy as jnp
import tensorflow.keras as tfk


def test_basic():
    y_true = jnp.array([[0.0, 1.0], [0.0, 0.0]])
    y_pred = jnp.array([[0.6, 0.4], [0.4, 0.6]])

    bce = elegy.losses.BinaryCrossentropy()
    result = bce(y_true, y_pred)
    assert jnp.isclose(result, 0.815, rtol=0.01)

    y_logits = jnp.log(y_pred) - jnp.log(1 - y_pred)
    bce = elegy.losses.BinaryCrossentropy(from_logits=True)
    result_from_logits = bce(y_true, y_logits)
    assert jnp.isclose(result_from_logits, 0.815, rtol=0.01)
    assert jnp.isclose(result_from_logits, result, rtol=0.01)

    bce = elegy.losses.BinaryCrossentropy()
    result = bce(y_true, y_pred, sample_weight=jnp.array([1, 0]))
    assert jnp.isclose(result, 0.458, rtol=0.01)

    bce = elegy.losses.BinaryCrossentropy(reduction=elegy.losses.Reduction.SUM)
    result = bce(y_true, y_pred)
    assert jnp.isclose(result, 1.630, rtol=0.01)

    bce = elegy.losses.BinaryCrossentropy(reduction=elegy.losses.Reduction.NONE)
    result = bce(y_true, y_pred)
    assert jnp.all(jnp.isclose(result, [0.916, 0.713], rtol=0.01))


#
def test_compatibility():

    # Input:  true (y_true) and predicted (y_pred) tensors
    y_true = jnp.array([[0.0, 1.0], [0.0, 0.0]])
    y_pred = jnp.array([[0.6, 0.4], [0.4, 0.6]])

    # Standard BCE, considering prediction tensor as probabilities
    bce_elegy = elegy.losses.BinaryCrossentropy()
    bce_tfk = tfk.losses.BinaryCrossentropy()
    assert jnp.isclose(bce_elegy(y_true, y_pred), bce_tfk(y_true, y_pred), rtol=0.0001)

    # Standard BCE, considering prediction tensor as logits
    y_logits = jnp.log(y_pred) - jnp.log(1 - y_pred)
    bce_elegy = elegy.losses.BinaryCrossentropy(from_logits=True)
    bce_tfk = tfk.losses.BinaryCrossentropy(from_logits=True)
    assert jnp.isclose(
        bce_elegy(y_true, y_logits), bce_tfk(y_true, y_logits), rtol=0.0001
    )

    # BCE using sample_weight
    bce_elegy = elegy.losses.BinaryCrossentropy()
    bce_tfk = tfk.losses.BinaryCrossentropy()
    assert jnp.isclose(
        bce_elegy(y_true, y_pred, sample_weight=jnp.array([1, 0])),
        bce_tfk(y_true, y_pred, sample_weight=jnp.array([1, 0])),
        rtol=0.0001,
    )

    # BCE with reduction method: SUM
    bce_elegy = elegy.losses.BinaryCrossentropy(reduction=elegy.losses.Reduction.SUM)
    bce_tfk = tfk.losses.BinaryCrossentropy(reduction=tfk.losses.Reduction.SUM)
    assert jnp.isclose(bce_elegy(y_true, y_pred), bce_tfk(y_true, y_pred), rtol=0.0001)

    # BCE with reduction method: NONE
    bce_elegy = elegy.losses.BinaryCrossentropy(reduction=elegy.losses.Reduction.NONE)
    bce_tfk = tfk.losses.BinaryCrossentropy(reduction=tfk.losses.Reduction.NONE)
    assert jnp.all(
        jnp.isclose(bce_elegy(y_true, y_pred), bce_tfk(y_true, y_pred), rtol=0.0001)
    )

    # BCE with label smoothing
    bce_elegy = elegy.losses.BinaryCrossentropy(label_smoothing=0.9)
    bce_tfk = tfk.losses.BinaryCrossentropy(label_smoothing=0.9)
    assert jnp.isclose(bce_elegy(y_true, y_pred), bce_tfk(y_true, y_pred), rtol=0.0001)



from dataclasses import dataclass
import sys
import typing as tp
from copy import copy
from enum import Enum
from functools import total_ordering

import jax
import jax.numpy as jnp
import jax.tree_util
import numpy as np


if sys.version_info >= (3, 8):
    from typing import Protocol, runtime_checkable
else:
    from typing_extensions import Protocol, runtime_checkable

EPSILON = 1e-7


class Mode(str, Enum):
    pred = "pred"
    test = "test"
    train = "train"
    summary = "summary"


class TrivialPytree:
    def tree_flatten(self):
        return tuple(vars(self).values()), None

    @classmethod
    def tree_unflatten(cls, _aux_data, children):
        return cls(*children)


class Empty:
    pass


EMPTY = Empty()


@jax.tree_util.register_pytree_node_class
class Uninitialized:
    def tree_flatten(self):
        return ((), None)

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        return cls()


UNINITIALIZED = Uninitialized()


@jax.tree_util.register_pytree_node_class
class RNGSeq(TrivialPytree):
    key: jnp.ndarray

    def __init__(self, key: tp.Union[int, jnp.ndarray]):
        self.key = (
            jax.random.PRNGKey(key) if isinstance(key, int) or key.shape == () else key
        )

    def next(self) -> jnp.ndarray:
        self.key = jax.random.split(self.key, 1)[0]
        return self.key

    def __repr__(self) -> str:
        return f"RNGSeq(key={self.key})"


T = tp.TypeVar("T")
Container = tp.Union[
    T,
    tp.Tuple["Container", ...],
    tp.Dict[str, "Container"],
]
ArrayHolder = tp.Union[Container[np.ndarray], np.ndarray]

IndexLike = tp.Union[str, int, tp.Iterable[tp.Union[str, int]]]

Shape = tp.Sequence[int]
ShapeLike = tp.Union[int, Shape]
FloatLike = tp.Union[float, np.ndarray]
DType = tp.Any
ParamName = str

Params = tp.Mapping[str, tp.Mapping[ParamName, np.ndarray]]
State = tp.Mapping[str, tp.Mapping[str, np.ndarray]]
PadFn = tp.Callable[[int], tp.Tuple[int, int]]
PadFnOrFns = tp.Union[PadFn, tp.Sequence[PadFn]]
PRNGKey = np.ndarray
Parameters = tp.Dict[str, tp.Any]
ParameterCollection = tp.Dict[str, Parameters]
Logs = tp.Dict[str, tp.Union[np.ndarray, float]]
Index = tp.Union[int, str]
Path = tp.Tuple[Index, ...]
Grads = tp.Any
RNG = tp.Union[RNGSeq, np.ndarray]
Scalar = tp.Union[np.ndarray, float, int]
SummaryModule = tp.Any
SummaryValue = tp.Any

NetParams = tp.Any
NetStates = tp.Any
ModuleParams = tp.Any
ModuleStates = tp.Any
MetricsStates = tp.Any
OptimizerStates = tp.Any
OptimizerStates = tp.Any
Grads = tp.Any
Pytree = tp.Any


@jax.tree_util.register_pytree_node_class
class Summary(tp.NamedTuple):
    path: Path
    module: tp.Optional[SummaryModule]
    value: SummaryValue

    def tree_flatten(self):
        return ((self.value,), (self.path, self.module))

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        (value,) = children
        path, module = aux_data

        return cls(path, module, value)


Summaries = tp.List[Summary]


class OutputStates(tp.NamedTuple):
    preds: tp.Any
    params: tp.Any
    states: tp.Any


class States(tp.NamedTuple):
    net_params: tp.Any = UNINITIALIZED
    net_states: tp.Any = UNINITIALIZED
    metrics_states: tp.Any = UNINITIALIZED
    optimizer_states: tp.Any = UNINITIALIZED
    rng: tp.Union[RNG, tp.Any] = UNINITIALIZED

    def update(
        self,
        net_params: tp.Any = UNINITIALIZED,
        net_states: tp.Any = UNINITIALIZED,
        metrics_states: tp.Any = UNINITIALIZED,
        optimizer_states: tp.Any = UNINITIALIZED,
        rng: tp.Union[RNG, Uninitialized] = UNINITIALIZED,
    ) -> "States":

        updates = {}

        if not isinstance(net_params, Uninitialized):
            updates["net_params"] = net_params
        if not isinstance(net_states, Uninitialized):
            updates["net_states"] = net_states
        if not isinstance(metrics_states, Uninitialized):
            updates["metrics_states"] = metrics_states
        if not isinstance(optimizer_states, Uninitialized):
            updates["optimizer_states"] = optimizer_states
        if not isinstance(rng, Uninitialized):
            updates["rng"] = rng

        kwargs = {field: getattr(self, field) for field in self._fields}
        kwargs.update(**updates)

        return States(**kwargs)

    def merge_new(self, other: "States") -> "States":

        updates = {}

        if isinstance(self.net_params, Uninitialized) and not isinstance(
            other.net_params, Uninitialized
        ):
            updates["net_params"] = other.net_params

        if isinstance(self.net_states, Uninitialized) and not isinstance(
            other.net_states, Uninitialized
        ):
            updates["net_states"] = other.net_states

        if isinstance(self.metrics_states, Uninitialized) and not isinstance(
            other.metrics_states, Uninitialized
        ):
            updates["metrics_states"] = other.metrics_states

        if isinstance(self.optimizer_states, Uninitialized) and not isinstance(
            other.optimizer_states, Uninitialized
        ):
            updates["optimizer_states"] = other.optimizer_states

        if isinstance(self.rng, Uninitialized) and not isinstance(
            other.rng, Uninitialized
        ):
            updates["rng"] = other.rng

        kwargs = {field: getattr(self, field) for field in self._fields}
        kwargs.update(**updates)

        return States(**kwargs)

    def merge(self, other: "States") -> "States":
        return other.update(*other)

    def copy(self) -> "States":
        return States(
            net_params=self.net_params,
            net_states=self.net_states,
            metrics_states=self.metrics_states,
            optimizer_states=self.optimizer_states,
            rng=copy(self.rng),
        )


@dataclass
class Parameter:
    collection: str
    value: tp.Any


@dataclass
class Info:
    shape: tp.Tuple[int, ...]
    dtype: tp.Any


@dataclass
class ParameterSpec:
    collection: str
    info: tp.Any


class Initializer(Protocol):
    def __call__(self, shape: Shape, dtype: DType) -> np.ndarray:
        ...


class JitCallable(Protocol):
    def __call__(self, *args) -> tp.Tuple[tp.Any, ParameterCollection]:
        ...


class InitJit(Protocol):
    def __call__(
        self,
        *,
        rng: tp.Optional[RNGSeq] = None,
        set_defaults: bool = False,
    ) -> JitCallable:
        ...


class ApplyJit(Protocol):
    def __call__(
        self,
        collections: ParameterCollection,
        *,
        training: bool = True,
        rng: tp.Optional[RNGSeq] = None,
        set_defaults: bool = False,
    ) -> JitCallable:
        ...


class MissingModule(Exception):
    pass


class MissingOptimizer(Exception):
    pass


class MissingMethod(Exception):
    pass


class DependencyUnavailable(Exception):
    pass


class ShapeMismatch(Exception):
    pass


class MissingParameter(Exception):
    pass


class NoContext(Exception):
    pass


class ModuleOrderError(Exception):
    pass


class SubmoduleNotRegistered(Exception):
    pass



import typing as tp
from abc import abstractmethod

import jax.numpy as jnp
from elegy import module, types, utils


class Metric(module.Module):
    """
    Encapsulates metric logic and state. Metrics accumulate state between `apply`s such
    that their output value reflect the metric as if calculated on the whole data
    given up to that point.

    Usage:

    ```python
    m = SomeMetric()
    _, state = m.init()(x)
    for x in batch:
        result = m.apply(state)(x)
    print('Final result: ', result)
    ```

    Usage with the Model API:

    ```python
    >>> import elegy, jax, optax
    >>> model = elegy.Model(
    ...     module=elegy.nn.Sequential(
    ...         lambda: [
    ...             elegy.nn.Flatten(),
    ...             elegy.nn.Linear(300),
    ...             jax.nn.relu,
    ...             elegy.nn.Linear(10),
    ...         ]
    ...     ),
    ...     loss=[
    ...         elegy.losses.SparseCategoricalCrossentropy(from_logits=True),
    ...     ],
    ...     metrics=[
    ...         elegy.metrics.SparseCategoricalAccuracy()
    ...     ],
    ...     optimizer=optax.rmsprop(1e-3),
    ... )

    ```

    To be implemented by subclasses:

    * `call()`: All state variables should be created in this method by
        calling `self.add_parameter(..., trainable=False)`, update this state by calling
        `self.update_parameter(...)`, and return a result based on these states.

    Example subclass implementation:

    ```python
    >>> class Accuracy(elegy.Metric):
    ...    def call(self, y_true, y_pred):
    ...
    ...        total = self.add_parameter("total", lambda: jnp.array(0), trainable=False)
    ...        count = self.add_parameter("count", lambda: jnp.array(0), trainable=False)
    ...
    ...        total += jnp.sum(y_true == y_pred)
    ...        count += jnp.prod(y_true.shape)
    ...
    ...        self.update_parameter("total", total)
    ...        self.update_parameter("count", count)
    ...
    ...        return total / count

    ```
    """

    __all__ = ["__init__", "call"]

    def __init__(self, on: tp.Optional[types.IndexLike] = None, **kwargs):
        """
        Base Metric constructor.

        Arguments:
            on: A string or integer, or iterable of string or integers, that
                indicate how to index/filter the `y_true` and `y_pred`
                arguments before passing them to `call`. For example if `on = "a"` then
                `y_true = y_true["a"]`. If `on` is an iterable
                the structures will be indexed iteratively, for example if `on = ["a", 0, "b"]`
                then `y_true = y_true["a"][0]["b"]`, same for `y_pred`. For more information
                check out [Keras-like behavior](https://poets-ai.github.io/elegy/guides/modules-losses-metrics/#keras-like-behavior).
        """

        super().__init__(**kwargs)

        self._labels_filter = (on,) if isinstance(on, (str, int)) else on

    def __call__(self, *args, **kwargs):

        if self._labels_filter is not None:
            if "y_true" in kwargs and kwargs["y_true"] is not None:
                for index in self._labels_filter:
                    kwargs["y_true"] = kwargs["y_true"][index]

            if "y_pred" in kwargs and kwargs["y_pred"] is not None:
                for index in self._labels_filter:
                    kwargs["y_pred"] = kwargs["y_pred"][index]

        return super().__call__(*args, **kwargs)

    @abstractmethod
    def call(self, *args, **kwargs):
        ...



import elegy


from elegy import utils, types
import jax.numpy as jnp
import jax
import tensorflow.keras as tfk

# import debugpy

# print("Waiting for debugger...")
# debugpy.listen(5679)
# debugpy.wait_for_client()


def test_basic():

    y_true = jnp.array([[0.0, 1.0], [0.0, 0.0]])
    y_pred = jnp.array([[1.0, 1.0], [1.0, 0.0]])

    # Using 'auto'/'sum_over_batch_size' reduction type.
    msle = elegy.losses.MeanSquaredLogarithmicError()

    assert msle(y_true, y_pred) == 0.24022643

    # Calling with 'sample_weight'.
    assert msle(y_true, y_pred, sample_weight=jnp.array([0.7, 0.3])) == 0.12011322

    # Using 'sum' reduction type.
    msle = elegy.losses.MeanSquaredLogarithmicError(
        reduction=elegy.losses.Reduction.SUM
    )

    assert msle(y_true, y_pred) == 0.48045287

    # Using 'none' reduction type.
    msle = elegy.losses.MeanSquaredLogarithmicError(
        reduction=elegy.losses.Reduction.NONE
    )

    assert jnp.equal(msle(y_true, y_pred), jnp.array([0.24022643, 0.24022643])).all()


def test_function():

    rng = jax.random.PRNGKey(42)

    y_true = jax.random.randint(rng, shape=(2, 3), minval=0, maxval=2)
    y_pred = jax.random.uniform(rng, shape=(2, 3))

    loss = elegy.losses.mean_squared_logarithmic_error(y_true, y_pred)

    assert loss.shape == (2,)

    first_log = jnp.log(jnp.maximum(y_true, types.EPSILON) + 1.0)
    second_log = jnp.log(jnp.maximum(y_pred, types.EPSILON) + 1.0)
    assert jnp.array_equal(loss, jnp.mean(jnp.square(first_log - second_log), axis=-1))


def test_compatibility():
    # Input:  true (y_true) and predicted (y_pred) tensors
    y_true = jnp.array([[0.0, 1.0], [0.0, 0.0]])
    y_pred = jnp.array([[0.6, 0.4], [0.4, 0.6]])

    # MSLE using sample_weight
    msle_elegy = elegy.losses.MeanSquaredLogarithmicError()
    msle_tfk = tfk.losses.MeanSquaredLogarithmicError()
    assert jnp.isclose(
        msle_elegy(y_true, y_pred, sample_weight=jnp.array([1, 0])),
        msle_tfk(y_true, y_pred, sample_weight=jnp.array([1, 0])),
        rtol=0.0001,
    )

    # MSLE with reduction method: SUM
    msle_elegy = elegy.losses.MeanSquaredLogarithmicError(
        reduction=elegy.losses.Reduction.SUM
    )
    msle_tfk = tfk.losses.MeanSquaredLogarithmicError(
        reduction=tfk.losses.Reduction.SUM
    )
    assert jnp.isclose(
        msle_elegy(y_true, y_pred), msle_tfk(y_true, y_pred), rtol=0.0001
    )

    # MSLE with reduction method: NONE
    msle_elegy = elegy.losses.MeanSquaredLogarithmicError(
        reduction=elegy.losses.Reduction.NONE
    )
    msle_tfk = tfk.losses.MeanSquaredLogarithmicError(
        reduction=tfk.losses.Reduction.NONE
    )
    assert jnp.all(
        jnp.isclose(msle_elegy(y_true, y_pred), msle_tfk(y_true, y_pred), rtol=0.0001)
    )


if __name__ == "__main__":

    test_basic()
    test_function()



import functools
import threading
import typing as tp
from abc import ABCMeta, abstractmethod
from contextlib import contextmanager
from dataclasses import dataclass
from copy import deepcopy

import jax
import jax.numpy as jnp
import numpy as np

from elegy import hooks, utils
from elegy import types

__all__ = [
    "Module",
    "to_module",
    "add_loss",
    "add_metric",
    "add_summary",
    "next_key",
]

T = tp.TypeVar("T")


class LocalContext(types.Protocol):
    parent: tp.Optional["Module"]
    module_path: tp.Optional[tp.Dict["Module", types.Path]]
    inside_call: tp.Optional[bool]
    module_index: tp.Optional[int]
    rng: tp.Optional[types.RNGSeq]
    training: tp.Optional[bool]
    initializing: tp.Optional[bool]


@dataclass
class _LocalContext(threading.local):
    parent: tp.Optional["Module"]
    module_path: tp.Optional[tp.Dict["Module", types.Path]]
    inside_call: tp.Optional[bool]
    module_index: tp.Optional[int]
    rng: tp.Optional[types.RNGSeq]
    training: tp.Optional[bool]
    initializing: tp.Optional[bool]


LOCAL: LocalContext = _LocalContext(
    parent=None,
    module_path=None,
    inside_call=None,
    module_index=None,
    rng=None,
    training=None,
    initializing=None,
)


def construct_module(cls, *args, **kwargs) -> "Module":
    # call constructor
    module: Module = cls.__new__(cls, *args, **kwargs)
    with instantiation_context(module):
        cls.__init__(module, *args, **kwargs)

    assert module is not None

    if not hasattr(module, "_submodules"):
        raise ValueError(
            "Constructing a Module without calling the super constructor "
            "is not supported."
        )

    # register submodules created in __init__
    for key, value in vars(module).items():
        if not key.startswith("_") and leaf_isinstance(value, Module):

            for path, submodule in utils.leaf_paths(value):
                if isinstance(submodule, Module):
                    path = (key,) + path
                    name = "/".join(map(str, path))
                    module._submodules[name] = submodule
                    module._submodule_name[submodule] = name

    return module


class ModuleMeta(ABCMeta):
    def __call__(cls: tp.Type, *args, **kwargs) -> "Module":

        # Set unique on parent when using inside `call`
        if LOCAL.inside_call:
            assert LOCAL.module_index is not None
            assert LOCAL.parent

            index = LOCAL.module_index
            parent = LOCAL.parent

            if len(parent._dynamic_submodules) > index:
                module = parent._dynamic_submodules[index]

                assert isinstance(module, Module)

                # if not isinstance(module, cls):
                if module.__class__.__name__ != cls.__name__:
                    raise types.ModuleOrderError(
                        f"Error retrieving module, expected type {cls.__name__}, got {module.__class__.__name__}. "
                        "This is probably due to control flow, you must guarantee that the same amount "
                        "of submodules will be created every time and that their order is the same."
                    )
            else:
                # if not LOCAL.initializing:
                #     raise ValueError(
                #         f"Trying to create module of type'{cls.__name__}' outside of `init`."
                #     )

                module = construct_module(cls, *args, **kwargs)

                name = utils.get_unique_name(set(parent._submodules), module.name)

                parent._submodules[name] = module
                parent._submodule_name[module] = name
                parent._dynamic_submodules.append(module)

            LOCAL.module_index += 1

            return module
        else:
            return construct_module(cls, *args, **kwargs)


class Module(metaclass=ModuleMeta):
    """
    Basic Elegy Module.

    For more information check out the [Module System guide](https://poets-ai.github.io/elegy/guides/module-system/).
    """

    name: str
    dtype: np.dtype
    _states_initial: tp.List[str]
    _submodules: tp.Dict[str, "Module"]
    _submodule_name: tp.Dict["Module", str]
    _dynamic_submodules: tp.List["Module"]
    _default_params: tp.Dict[str, types.Parameter]
    _scope_params: tp.Optional[tp.Dict[str, types.Parameter]]
    _spec: tp.Dict[str, types.ParameterSpec]
    _initialized: bool = False

    init_jit: types.InitJit
    apply_jit: types.ApplyJit

    __all__ = [
        "__init__",
        "call",
        "add_parameter",
        "get_parameters",
        "set_parameters",
        "reset",
        "init",
    ]

    def __init__(self, name: tp.Optional[str] = None, dtype: tp.Any = jnp.float32):
        """
        Initializes the current module with the given name.

        Subclasses should call this constructor before creating other modules or
        variables such that those modules are named correctly.

        Arguments:
            name: An optional string name for the class. If ``name`` is not provided then the class name for the
                current instance is converted to ``lower_snake_case`` and used instead.
        """
        self.name = name if name else utils.lower_snake_case(self.__class__.__name__)
        self.dtype = dtype
        self._submodules = {}
        self._submodule_name = {}
        self._dynamic_submodules = []
        self._default_params = {}
        self._scope_params = None
        self._spec = {}
        self._initialized = False

        self._signature_f = self.call

        self._jit_functions()

    @property
    def initialized(self):
        return self._initialized

    def _mark_initialized_recursive(self):
        self._initialized = True
        for submodule in self._submodules.values():
            submodule._mark_initialized_recursive()

    def _register_parameter(self, name: str, parameter: types.Parameter):
        param_info = jax.tree_map(
            lambda x: types.Info(shape=x.shape, dtype=x.dtype),
            parameter.value,
        )
        self._spec[name] = types.ParameterSpec(
            collection=parameter.collection, info=param_info
        )
        # commenting this out makes Module.init stateful by default
        # self._default_params[name] = parameter

    def call_with_defaults(
        self,
        *,
        rng: tp.Optional[types.RNGSeq] = None,
        training: bool = False,
    ) -> tp.Callable[..., tp.Any]:
        def call_with_defaults_wrapper(*args, **kwargs):
            if not self.initialized:
                _, collections = self.init(rng=rng, set_defaults=True)(*args, **kwargs)
            else:
                collections = self.get_default_parameters()

            y, _ = self.apply(
                collections,
                training=training,
                rng=rng,
                set_defaults=True,
            )(*args, **kwargs)

            return y

        return call_with_defaults_wrapper

    def call_with_defaults_jit(
        self,
        *,
        rng: tp.Optional[types.RNGSeq] = None,
        training: bool = False,
    ) -> tp.Callable[..., tp.Any]:
        def call_with_defaults_jit_wrapper(*args):

            if not self.initialized:
                _, collections = self.init_jit(rng=rng)(*args)
            else:
                collections = self.get_default_parameters()

            y, collections = self.apply_jit(
                collections,
                training=training,
                rng=rng,
            )(*args)

            self.set_default_parameters(collections)

            return y

        return call_with_defaults_jit_wrapper

    def _jit_functions(self):
        # ------------------------------
        # init
        # ------------------------------
        def init_jit_raw(rng: tp.Optional[types.RNGSeq], *args):
            return self.init(rng=rng)(*args)

        init_jit: types.JitCallable = hooks.jit(init_jit_raw)

        def init_jit_wrapper(
            rng: tp.Optional[types.RNGSeq] = None,
            set_defaults: bool = False,
        ) -> types.JitCallable:
            def init_jit_callable(
                *args,
            ) -> tp.Tuple[tp.Any, types.ParameterCollection]:

                y, collections = init_jit(rng, *args)

                if set_defaults:
                    self.set_default_parameters(collections)

                return y, collections

            init_jit_callable._signature_f = self.call

            return init_jit_callable

        # ------------------------------
        # apply
        # ------------------------------
        def apply_jit_raw(
            parameters, training: bool, rng: tp.Optional[types.RNGSeq], *args
        ):
            return self.apply(
                collections=parameters,
                training=training,
                rng=rng,
            )(*args)

        apply_jit = hooks.jit(apply_jit_raw, static_argnums=[1])

        def apply_jit_wrapper(
            collections: types.ParameterCollection,
            *,
            training: bool = True,
            rng: tp.Optional[types.RNGSeq] = None,
            set_defaults: bool = False,
        ) -> types.JitCallable:
            def apply_jit_callable(
                *args,
            ) -> tp.Tuple[tp.Any, types.ParameterCollection]:

                y, collections_ = apply_jit(collections, training, rng, *args)

                if set_defaults:
                    self.set_default_parameters(collections_)

                return y, collections_

            apply_jit_callable._signature_f = self.call

            return apply_jit_callable

        # ------------------------------
        # assign functions
        # ------------------------------
        self.init_jit = init_jit_wrapper
        self.apply_jit = apply_jit_wrapper

    def __setstate__(self, d):
        self.__dict__ = d
        self._jit_functions()

    def __getstate__(self):
        d = dict(self.__dict__)
        del d["apply_jit"]
        del d["init_jit"]

        d["_default_params"] = {}
        d["_scope_params"] = None

        return d

    def __call__(self, *args, **kwargs) -> tp.Any:
        """
        Forwards all input arguments to the Module's `call` method and calls
        `elegy.hooks.add_summary` on the outputs.
        """

        # this marks initialization

        with call_context(self):

            outputs = self.call(*args, **kwargs)

            if hooks.summaries_active():
                path = get_module_path(self)
                assert path is not None
                hooks.add_summary(path, self, outputs)

            return outputs

    @abstractmethod
    def call(self, *args, **kwargs):
        ...

    def add_summary(self, name: str, f: tp.Any, value: tp.Any):
        if hooks.summaries_active():
            path = get_module_path(self) + (name,)
            assert path is not None
            hooks.add_summary(path, f, value)

    def init(
        self,
        *,
        rng: tp.Optional[types.RNGSeq] = None,
        set_defaults: bool = False,
    ) -> tp.Callable[..., tp.Tuple[tp.Any, types.ParameterCollection]]:
        """
        Initializes the module,
        """

        def init_callable(
            *args, **kwargs
        ) -> tp.Tuple[tp.Any, types.ParameterCollection]:

            with module_context(
                module=self, collections=None, initializing=True, training=True, rng=rng
            ):
                y = self(*args, **kwargs)
                collections = self.get_parameters_internal()

            self._mark_initialized_recursive()

            if set_defaults:
                self.set_default_parameters(collections)

            return y, collections

        init_callable._signature_f = self.call

        return init_callable

    def apply(
        self,
        collections: types.ParameterCollection,
        *,
        training: bool = True,
        rng: tp.Optional[types.RNGSeq] = None,
        set_defaults: bool = False,
    ) -> tp.Callable[
        ..., tp.Union[tp.Any, tp.Tuple[tp.Any, types.ParameterCollection]]
    ]:
        """
        Call the module.
        """

        def apply_callable(
            *args, **kwargs
        ) -> tp.Tuple[tp.Any, types.ParameterCollection]:

            with module_context(
                module=self,
                collections=collections,
                initializing=False,
                training=training,
                rng=rng,
            ):
                y = self(*args, **kwargs)
                collections_ = self.get_parameters_internal()

            if set_defaults:
                self.set_default_parameters(collections_)

            return y, collections_

        apply_callable._signature_f = self.call

        return apply_callable

    def __getattr__(self, name: str) -> tp.Any:

        if name in self._submodules:
            return self._submodules[name]

        if self._scope_params is not None and name in self._scope_params:
            return self._scope_params[name].value

        if not has_scope() and self.initialized and name in self._default_params:
            return self._default_params[name].value

        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute '{name}'"
        )

    def add_parameter(
        self,
        name: str,
        initializer: tp.Callable[[], tp.Any],
        collection: tp.Optional[str] = None,
        trainable: bool = True,
        regularizer: tp.Optional[tp.Callable[[tp.Any], jnp.ndarray]] = None,
        constraint: tp.Optional[tp.Callable[[tp.Any], tp.Any]] = None,
    ) -> tp.Any:
        """
        Adds a parameter to the current module. The parameter will only be initialized once and
        will reused afterwards.

        Arguments:
            name: The name of the parameter. It must be unique and no other field/property/method
                of the instance can have that name.
            initializer: A callable that takes not arguments returns the initial value.
            collection: Optional name of the parameter collection, if not defined it will be se to
                `"parameters"` if `trainable=True` else it will be set to `"states"`.
            trainable: Specify whether this parameter should be added to the default trainable `"parameters"`
                collection or to the default non-trainable `"states"` collection. If collection is
                passed this parameter will ignored.
            regularizer: Regularizer instance (callable).
            constraint: Constraint instance (callable).

        Returns:
            The value of the parameter.
        """

        if collection is None:
            collection = "parameters" if trainable else "states"

        if self._scope_params is None:
            self._scope_params = {}

        if name not in self._scope_params:

            if not self.is_initializing():
                raise ValueError(f"Cannot add parameter {name} outside `init`")

            if name in self._default_params:
                parameter = self._default_params[name]
                assert collection == parameter.collection
                initial_value = parameter.value

            else:
                initial_value = initializer()
                initial_value = jax.tree_map(jnp.asarray, initial_value)
                parameter = types.Parameter(collection, initial_value)
                self._register_parameter(name, parameter)

            self._scope_params[name] = parameter

        parameter = self._scope_params[name]

        if parameter.collection != collection:
            raise ValueError(
                f"types.Parameter {name} previously found in collection {parameter.collection} "
                f"but currently being added for collection {collection}"
            )

        value = parameter.value

        if constraint is not None:
            value = constraint(value)

        if regularizer is not None:
            hooks.add_loss(
                name=utils.get_name(regularizer),
                value=regularizer(value),
            )

        return value

    def update_parameter(self, name: str, value: tp.Any) -> None:
        """
        Update a parameter of the current module.

        !!! Note
            types.Parameters are not updated when `Module.init` is called.

        Arguments:
            name: The name of the parameter to be updated. It must be unique and no other field/property/method
                of the instance can have that name.
            value: The updated value of the state.

        Raises:
            `ValueError` if parameter is not present in current module.
        """

        assert self._scope_params is not None

        if name not in self._scope_params:
            raise ValueError(f"types.Parameter {name} not found in {self}.")

        if self.is_initializing():
            return

        parameter = self._scope_params[name]
        assert isinstance(parameter, types.Parameter)
        parameter.value = value

    def add_or_update_parameter(
        self,
        name: str,
        value: tp.Callable[[], tp.Any],
        collection: tp.Optional[str] = None,
        trainable: bool = True,
    ):
        """
        Add a parameter to the current module or update it if it already exists.

        !!! Note
            types.Parameters are not updated when `Module.init` is called.

        Arguments:
            name: The name of the state. It must be unique and no other field/property/method
                of the instance can have that name.
            value: The updated value of the state.
            collection: Optional name of the parameter collection, if not defined it will be se to
                `"parameters"` if `trainable=True` else it will be set to `"states"`.
            trainable: Specify whether this parameter should be added to the default trainable `"parameters"`
                collection or to the default non-trainable `"states"` collection. If collection is
                passed this parameter will ignored.

        Raises:
            `ValueError` if parameter is not present in current module.
        """
        assert self._scope_params is not None

        if name not in self._scope_params:
            self.add_parameter(
                name, lambda: value, trainable=trainable, collection=collection
            )
        else:
            self.update_parameter(name, value)

    def next_key(self) -> jnp.ndarray:
        if LOCAL.rng is None:
            raise types.NoContext(
                f"Trying to call `next_key` from module {self} outside `init` or `apply`."
            )

        return LOCAL.rng.next()

    def is_training(self) -> bool:
        if LOCAL.training is None:
            raise types.NoContext(
                f"Trying to call `is_training` from module '{self}' outside `init` or `apply`."
            )
        return LOCAL.training

    def is_initializing(self) -> bool:
        if LOCAL.initializing is None:
            raise types.NoContext(
                f"Trying to call `is_initializing` from module '{self}' outside `init` or `apply`."
            )

        return LOCAL.initializing

    def clear_default_parameters(self):
        self._clear_parameters(default_params=True)

    def _clear_parameters(self, default_params: bool = False):

        self._scope_params = None

        if default_params:
            self._default_params.clear()

        for submodule in self._submodules.values():
            submodule._clear_parameters(default_params=default_params)

    def set_default_parameters(
        self,
        collections: types.ParameterCollection,
    ):
        self._set_parameters_internal(
            collections=collections,
            set_default_params=True,
        )

    def _set_parameters_internal(
        self,
        collections: types.ParameterCollection,
        set_default_params: bool = False,
    ):
        old_colletions = self.get_default_parameters()
        collections = jax.tree_map(jnp.asarray, collections)

        try:
            self._set_parameters(
                collections=collections,
                set_default_params=set_default_params,
            )
        except:
            self._set_parameters(
                collections=old_colletions,
                set_default_params=set_default_params,
            )
            raise

    def _validate_parameters(self, collections: types.ParameterCollection):
        # define sets
        incoming_names = set(
            name for parameters in collections.values() for name in parameters
        )
        expected_submodule_names = set(self._submodules)
        expected_parameter_names = set(self._spec)

        # check missing parameters
        missing_parameters = expected_parameter_names - incoming_names
        if missing_parameters:
            raise ValueError(
                f"Missing parameters {missing_parameters} for module {self.name}"
            )

        # check unkown names
        unknown_parameters = (
            incoming_names - expected_submodule_names - expected_parameter_names
        )
        if unknown_parameters:
            raise ValueError(
                f"Got unknown parameters {unknown_parameters} on module {self.name}"
            )

        for name, param_spec in self._spec.items():
            parameter = utils.get_parameter(collections, name)

            if parameter.collection != param_spec.collection:
                raise ValueError(
                    f"types.Parameter {name} on module {self.name} was expected to be on collection {param_spec.collection} "
                    f"but was found on collection {parameter.collection}"
                )

            def validate_value(value: jnp.ndarray, info: types.Info):
                if value.shape != info.shape:
                    incoming_shapes = jax.tree_map(lambda x: x.shape, parameter.value)
                    expected_shape = jax.tree_map(lambda x: x.shape, param_spec.info)

                    raise ValueError(
                        f"Shape mismatch in parameter {name} on module {self.name}.\n"
                        f"Got: {incoming_shapes}\n"
                        f"Expected: {expected_shape}"
                    )

            jax.tree_multimap(validate_value, parameter.value, param_spec.info)

    def _set_parameters(
        self,
        collections: types.ParameterCollection,
        check_shapes: bool = False,
        set_default_params: bool = False,
    ):
        if not self.initialized:
            raise ValueError(f"Cannot set parameters for uninitialized module {self}")
        assert self._spec is not None

        self._validate_parameters(collections)

        # build new parameters
        new_params = {
            name: utils.get_parameter(collections, name) for name in self._spec
        }

        # set module parameters
        if set_default_params:
            self._default_params = new_params
        else:
            self._scope_params = new_params

        # set submodule parameters
        for name, submodule in self._submodules.items():
            submodule._set_parameters(
                collections=utils.get_submodule_colletions(collections, name),
                check_shapes=check_shapes,
                set_default_params=set_default_params,
            )

    def get_default_parameters(self) -> types.ParameterCollection:
        parameters = self._get_parameters(defaults=True)
        return utils.split_into_collections(parameters)

    def get_parameters_internal(
        self, defaults: bool = False
    ) -> types.ParameterCollection:
        parameters = self._get_parameters(defaults=defaults)
        return utils.split_into_collections(parameters)

    def _get_parameters(self, defaults: bool) -> tp.Dict[str, tp.Any]:
        # if (self._params is None and not defaults) or (
        #     not self.initialized and defaults
        # ):
        #     return {}

        parameters: tp.Dict[str, tp.Any]

        if defaults:
            if self.initialized:
                parameters = self._default_params.copy()
            else:
                raise ValueError(
                    f"Cannot get default parameters from uninitialized Module {self.name}"
                )
        else:
            if self._scope_params is not None:
                parameters = self._scope_params.copy()
            else:
                parameters = {}

        for name, submodule in self._submodules.items():
            parameters[name] = submodule._get_parameters(defaults=defaults)

        return parameters

    def has_parameter(self, name: str) -> bool:
        return hasattr(self, name)


# -------------------------------------------------------------
# hooks
# -------------------------------------------------------------


def next_key() -> jnp.ndarray:
    if LOCAL.rng is None:
        raise ValueError(f"No rng present in context, please set it in `context`.")

    return LOCAL.rng.next()


def get_module_path(module: tp.Optional[Module] = None) -> tp.Optional[types.Path]:
    if module is None:
        module = LOCAL.parent

    return LOCAL.module_path[module] if module is not None and has_scope() else None


def get_module_path_str(module: Module) -> tp.Optional[str]:
    return (
        "/".join(map(str, LOCAL.module_path[module]))
        if module is not None and has_scope()
        else None
    )


def has_scope() -> bool:
    return LOCAL.module_path is not None


# -----------------------------------------------------------------------------
# context managers
# -----------------------------------------------------------------------------


def module_context(
    module: Module,
    collections: tp.Optional[types.ParameterCollection],
    initializing: bool,
    training: bool,
    rng: tp.Optional[types.RNGSeq],
) -> tp.ContextManager[None]:
    return _scope_context(
        module=module,
        collections=collections,
        initializing=initializing,
        training=training,
        rng=rng,
    )


@contextmanager
def _scope_context(
    module: Module,
    collections: tp.Optional[types.ParameterCollection],
    initializing: bool,
    training: bool,
    rng: tp.Optional[types.RNGSeq],
):
    prev_module_path = LOCAL.module_path
    prev_initializing = LOCAL.initializing
    prev_training = LOCAL.training
    prev_rng = LOCAL.rng

    LOCAL.module_path = {}
    LOCAL.initializing = initializing
    LOCAL.training = training
    LOCAL.rng = rng

    try:
        if collections is not None:
            module._set_parameters_internal(collections)

        yield
    finally:
        module._clear_parameters()
        LOCAL.module_path = prev_module_path
        LOCAL.initializing = prev_initializing
        LOCAL.training = prev_training
        LOCAL.rng = prev_rng


def call_context(module: Module) -> tp.ContextManager[None]:
    return _call_context(module)


@contextmanager
def _call_context(module: Module):

    prev_parent = LOCAL.parent
    prev_inside_call = LOCAL.inside_call
    prev_module_index = LOCAL.module_index

    LOCAL.parent = module
    LOCAL.inside_call = True
    LOCAL.module_index = 0

    try:

        if not has_scope():
            raise types.NoContext(
                f"Trying to call top-level module '{module}' directly, use `apply` instead."
            )

        elif prev_parent is not None and module not in prev_parent._submodule_name:
            raise types.SubmoduleNotRegistered(
                f"Submodule {utils.get_name(module)} not registered in {utils.get_name(prev_parent)}, "
                f"this is probably due to some of the following reasons:\n"
                f"- The submodule is being captured by closure and not registered to any field.\n"
                f"- The submodule was set to a field of the parent but "
                f"its contained inside a more complex type which elegy cannot "
                f"inspect, elegy only looks structures of (possibly nested) list, tuple, or dict.\n"
                f"- The submodule was set to a field of the parent by mutating such field after __init__\n\n"
                f"- If non of the previous is true consider this a bug."
                f"Submodule: {module}\n"
                f"Module: {prev_parent}\n"
            )

        if prev_parent is None:
            LOCAL.module_path[module] = ()
        else:
            parent_path = LOCAL.module_path[prev_parent]
            child_name = prev_parent._submodule_name[module]
            LOCAL.module_path[module] = parent_path + (child_name,)
        yield
    finally:
        LOCAL.parent = prev_parent
        LOCAL.inside_call = prev_inside_call
        LOCAL.module_index = prev_module_index


def instantiation_context(module: Module) -> tp.ContextManager[None]:
    return _instantiation_context(module)


@contextmanager
def _instantiation_context(module: Module):

    prev_inside_call = LOCAL.inside_call
    prev_module_path = LOCAL.module_path

    LOCAL.inside_call = False
    LOCAL.module_path = None

    try:
        yield
    finally:
        LOCAL.inside_call = prev_inside_call
        LOCAL.module_path = prev_module_path


# ------------------------------------------------------------------------
# utils
# ------------------------------------------------------------------------


def leaf_isinstance(obj: tp.Any, types) -> bool:

    if isinstance(obj, (tp.List, tp.Tuple)) and obj:
        return any(leaf_isinstance(elem, types) for elem in obj)
    elif isinstance(obj, tp.Dict) and obj:
        return any(leaf_isinstance(elem, types) for elem in obj.values())
    else:
        return isinstance(obj, types)


def to_module(f):
    class ToModule(Module):
        def __init__(self, name: tp.Optional[str] = None):
            super().__init__(
                name=utils.lower_snake_case(f.__name__) if name is None else name
            )
            self._signature_f = f
            self.f = f

        def call(self, *args, **kwargs):
            return self.f(*args, **kwargs)

    ToModule.__name__ = f.__name__

    return ToModule



import typing as tp

import jax
import jax.numpy as jnp
import numpy as np

from elegy.module import Module
from elegy.nn.dropout import Dropout
from elegy.nn.layer_normalization import LayerNormalization
from elegy.nn.linear import Linear
from elegy.nn.multi_head_attention import MultiHeadAttention
from elegy.nn.sequential_module import sequential


class TransformerEncoderLayer(Module):
    r"""
    TransformerEncoderLayer is made up of self-attn and feedforward network.

    This standard encoder layer is based on the paper "Attention Is All You Need".
    Ashish Vaswani, Noam Shazeer, Niki Parmar, Jakob Uszkoreit, Llion Jones, Aidan N Gomez,
    Lukasz Kaiser, and Illia Polosukhin. 2017. Attention is all you need. In Advances in
    Neural Information Processing Systems, pages 6000-6010. Users may modify or implement
    in a different way during application.

    Args:
        head_size: the number of expected features in the input (required).
        num_heads: the number of heads in the multiheadattention models (required).
        output_size: the dimension of the feedforward network model (default=2048).
        dropout: the dropout value (default=0.1).
        activation: the activation function of intermediate layer, relu or gelu (default=relu).

    Examples::
        >>> # encoder_layer = nn.TransformerEncoderLayer(head_size=512, num_heads=8)
        >>> # src = torch.rand(10, 32, 512)
        >>> # out = encoder_layer(src)
    """

    def __init__(
        self,
        head_size: int,
        num_heads: int,
        output_size: tp.Optional[int] = None,
        dropout: float = 0.0,
        activation=jax.nn.relu,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.head_size = head_size
        self.num_heads = num_heads
        self.output_size = output_size
        self.dropout = dropout
        self.activation = activation

    def call(
        self,
        src: np.ndarray,
        mask: tp.Optional[np.ndarray] = None,
        # src_key_padding_mask: tp.Optional[np.ndarray] = None,
    ) -> np.ndarray:
        r"""Pass the input through the encoder layer.

        Args:
            src: the sequence to the encoder layer (required).
            mask: the mask for the src sequence (optional).
            src_key_padding_mask: the mask for the src keys per batch (optional).

        Shape:
            see the docs in Transformer class.
        """
        # Implementation of Feedforward model

        output_size: int = (
            self.output_size if self.output_size is not None else src.shape[-1]
        )

        src2 = MultiHeadAttention(
            self.head_size,
            self.num_heads,
            dropout=self.dropout,
        )(src, mask=mask)
        src = src + Dropout(self.dropout)(src2)
        src = LayerNormalization()(src)
        src2 = sequential(
            Linear(output_size),
            self.activation,
            Dropout(self.dropout),
            Linear(output_size),
        )(src)
        src = src + Dropout(self.dropout)(src2)
        src = LayerNormalization()(src)
        return src


class TransformerEncoder(Module):
    r"""
    TransformerEncoder is a stack of N encoder layers

    Args:
        encoder_layer: an instance of the TransformerEncoderLayer() class (required).
        num_layers: the number of sub-encoder-layers in the encoder (required).
        norm: the layer normalization component (optional).

    Examples::
        >>> # transformer_encoder = elegy.nn.TransformerEncoder(
                lambda: elegy.nn.TransformerEncoderLayer(head_size=512, num_heads=8),
                num_layers=6,
            )
        >>> # src = torch.rand(10, 32, 512)
        >>> # out = transformer_encoder(src)
    """

    def __init__(
        self,
        encoder_layer: tp.Callable[[], Module],
        num_layers: int,
        norm: tp.Optional[tp.Callable[[], Module]] = None,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.encoder_layer = encoder_layer
        self.num_layers = num_layers
        self.norm = norm

    def call(
        self,
        src: np.ndarray,
        mask: tp.Optional[np.ndarray] = None,
        # src_key_padding_mask: tp.Optional[np.ndarray] = None,
    ) -> np.ndarray:
        r"""Pass the input through the encoder layers in turn.

        Args:
            src: the sequence to the encoder (required).
            mask: the mask for the src sequence (optional).
            src_key_padding_mask: the mask for the src keys per batch (optional).

        Shape:
            see the docs in Transformer class.
        """
        output = src

        for _ in range(self.num_layers):
            output = self.encoder_layer()(output, mask=mask)

        if self.norm is not None:
            output = self.norm()(output)

        return output


class TransformerDecoderLayer(Module):
    r"""TransformerDecoderLayer is made up of self-attn, multi-head-attn and feedforward network.
    This standard decoder layer is based on the paper "Attention Is All You Need".
    Ashish Vaswani, Noam Shazeer, Niki Parmar, Jakob Uszkoreit, Llion Jones, Aidan N Gomez,
    Lukasz Kaiser, and Illia Polosukhin. 2017. Attention is all you need. In Advances in
    Neural Information Processing Systems, pages 6000-6010. Users may modify or implement
    in a different way during application.

    Args:
        head_size: the number of expected features in the input (required).
        num_heads: the number of heads in the multiheadattention models (required).
        output_size: the dimension of the feedforward network model (default=2048).
        dropout: the dropout value (default=0.1).
        activation: the activation function of intermediate layer, relu or gelu (default=relu).

    Examples::
        >>> # decoder_layer = nn.TransformerDecoderLayer(head_size=512, num_heads=8)
        >>> # memory = torch.rand(10, 32, 512)
        >>> # tgt = torch.rand(20, 32, 512)
        >>> # out = decoder_layer(tgt, memory)
    """

    def __init__(
        self,
        head_size: int,
        num_heads: int,
        output_size: int = 2048,
        dropout: float = 0.1,
        activation: tp.Callable[[np.ndarray], np.ndarray] = jax.nn.relu,
    ):
        super().__init__()
        self.head_size = head_size
        self.num_heads = num_heads
        self.output_size = output_size
        self.dropout = dropout
        self.activation = activation

    def call(
        self,
        tgt: np.ndarray,
        memory: np.ndarray,
        tgt_mask: tp.Optional[np.ndarray] = None,
        memory_mask: tp.Optional[np.ndarray] = None,
        # tgt_key_padding_mask: tp.Optional[np.ndarray] = None,
        # memory_key_padding_mask: tp.Optional[np.ndarray] = None,
    ) -> np.ndarray:
        r"""Pass the inputs (and mask) through the decoder layer.

        Args:
            tgt: the sequence to the decoder layer (required).
            memory: the sequence from the last layer of the encoder (required).
            tgt_mask: the mask for the tgt sequence (optional).
            memory_mask: the mask for the memory sequence (optional).
            tgt_key_padding_mask: the mask for the tgt keys per batch (optional).
            memory_key_padding_mask: the mask for the memory keys per batch (optional).

        Shape:
            see the docs in Transformer class.
        """
        # Implementation of Feedforward model

        tgt2 = MultiHeadAttention(self.head_size, self.num_heads, dropout=self.dropout)(
            tgt, mask=tgt_mask
        )
        tgt = tgt + Dropout(self.dropout)(tgt2)
        tgt = LayerNormalization()(tgt)
        tgt2 = MultiHeadAttention(self.head_size, self.num_heads, dropout=self.dropout)(
            tgt,
            memory,
            mask=memory_mask,
        )
        tgt = tgt + Dropout(self.dropout)(tgt2)
        tgt = LayerNormalization()(tgt)
        tgt = tgt + sequential(
            Linear(self.output_size),
            self.activation,
            Dropout(self.dropout),
            Linear(self.output_size),
            Dropout(self.dropout),
        )(tgt)
        tgt = LayerNormalization()(tgt)
        return tgt


class TransformerDecoder(Module):
    r"""TransformerDecoder is a stack of N decoder layers

    Args:
        decoder_layer: an instance of the TransformerDecoderLayer() class (required).
        num_layers: the number of sub-decoder-layers in the decoder (required).
        norm: the layer normalization component (optional).

    Examples::
        >>> # decoder_layer = nn.TransformerDecoderLayer(head_size=512, num_heads=8)
        >>> # transformer_decoder = nn.TransformerDecoder(decoder_layer, num_layers=6)
        >>> # memory = torch.rand(10, 32, 512)
        >>> # tgt = torch.rand(20, 32, 512)
        >>> # out = transformer_decoder(tgt, memory)
    """

    def __init__(
        self,
        decoder_layer: tp.Callable[[], Module],
        num_layers: int,
        norm: tp.Optional[tp.Callable[[], Module]] = None,
    ):
        super().__init__()
        self.decoder_layer = decoder_layer
        self.num_layers = num_layers
        self.norm = norm

    def call(
        self,
        tgt: np.ndarray,
        memory: np.ndarray,
        tgt_mask: tp.Optional[np.ndarray] = None,
        memory_mask: tp.Optional[np.ndarray] = None,
        # tgt_key_padding_mask: tp.Optional[np.ndarray] = None,
        # memory_key_padding_mask: tp.Optional[np.ndarray] = None,
    ) -> np.ndarray:
        r"""Pass the inputs (and mask) through the decoder layer in turn.

        Args:
            tgt: the sequence to the decoder (required).
            memory: the sequence from the last layer of the encoder (required).
            tgt_mask: the mask for the tgt sequence (optional).
            memory_mask: the mask for the memory sequence (optional).
            tgt_key_padding_mask: the mask for the tgt keys per batch (optional).
            memory_key_padding_mask: the mask for the memory keys per batch (optional).

        Shape:
            see the docs in Transformer class.
        """
        output = tgt

        for _ in range(self.num_layers):
            output = self.decoder_layer()(
                output,
                memory,
                tgt_mask=tgt_mask,
                memory_mask=memory_mask,
                # tgt_key_padding_mask=tgt_key_padding_mask,
                # memory_key_padding_mask=memory_key_padding_mask,
            )

        if self.norm is not None:
            output = self.norm()(output)

        return output


class Transformer(Module):
    r"""A transformer model. User is able to modify the attributes as needed. The architecture
    is based on the paper "Attention Is All You Need". Ashish Vaswani, Noam Shazeer,
    Niki Parmar, Jakob Uszkoreit, Llion Jones, Aidan N Gomez, Lukasz Kaiser, and
    Illia Polosukhin. 2017. Attention is all you need. In Advances in Neural Information
    Processing Systems, pages 6000-6010. Users can build the BERT(https://arxiv.org/abs/1810.04805)
    model with corresponding parameters.

    Args:
        head_size: the number of expected features in the encoder/decoder inputs (default=512).
        num_heads: the number of heads in the multiheadattention models (default=8).
        num_encoder_layers: the number of sub-encoder-layers in the encoder (default=6).
        num_decoder_layers: the number of sub-decoder-layers in the decoder (default=6).
        output_size: the dimension of the feedforward network model (default=2048).
        dropout: the dropout value (default=0.1).
        activation: the activation function of encoder/decoder intermediate layer, relu or gelu (default=relu).
        custom_encoder: custom encoder (default=None).
        custom_decoder: custom decoder (default=None).

    Examples::
        >>> # transformer_model = nn.Transformer(num_heads=16, num_encoder_layers=12)
        >>> # src = torch.rand((10, 32, 512))
        >>> # tgt = torch.rand((20, 32, 512))
        >>> # out = transformer_model(src, tgt)

    Note: A full example to apply nn.Transformer module for the word language model is available in
    https://github.com/pytorch/examples/tree/master/word_language_model
    """

    def __init__(
        self,
        head_size: int = 512,
        num_heads: int = 8,
        num_encoder_layers: int = 6,
        num_decoder_layers: int = 6,
        output_size: int = 2048,
        dropout: float = 0.1,
        activation: tp.Callable[[np.ndarray], np.ndarray] = jax.nn.relu,
        custom_encoder: tp.Optional[tp.Any] = None,
        custom_decoder: tp.Optional[tp.Any] = None,
        **kwargs
    ) -> None:
        super().__init__(**kwargs)
        self.head_size = head_size
        self.num_heads = num_heads
        self.num_encoder_layers = num_encoder_layers
        self.num_decoder_layers = num_decoder_layers
        self.output_size = output_size
        self.dropout = dropout
        self.activation = activation
        self.custom_encoder = custom_encoder
        self.custom_decoder = custom_decoder

    def call(
        self,
        src: np.ndarray,
        tgt: np.ndarray,
        src_mask: tp.Optional[np.ndarray] = None,
        tgt_mask: tp.Optional[np.ndarray] = None,
        memory_mask: tp.Optional[np.ndarray] = None,
        # src_key_padding_mask: tp.Optional[np.ndarray] = None,
        # tgt_key_padding_mask: tp.Optional[np.ndarray] = None,
        # memory_key_padding_mask: tp.Optional[np.ndarray] = None,
    ) -> np.ndarray:
        r"""Take in and process masked source/target sequences.

        Args:
            src: the sequence to the encoder (required).
            tgt: the sequence to the decoder (required).
            src_mask: the additive mask for the src sequence (optional).
            tgt_mask: the additive mask for the tgt sequence (optional).
            memory_mask: the additive mask for the encoder output (optional).
            src_key_padding_mask: the ByteTensor mask for src keys per batch (optional).
            tgt_key_padding_mask: the ByteTensor mask for tgt keys per batch (optional).
            memory_key_padding_mask: the ByteTensor mask for memory keys per batch (optional).

        Shape:
            - src: :math:`(S, N, E)`.
            - tgt: :math:`(T, N, E)`.
            - src_mask: :math:`(S, S)`.
            - tgt_mask: :math:`(T, T)`.
            - memory_mask: :math:`(T, S)`.
            - src_key_padding_mask: :math:`(N, S)`.
            - tgt_key_padding_mask: :math:`(N, T)`.
            - memory_key_padding_mask: :math:`(N, S)`.

            Note: [src/tgt/memory]_mask ensures that position i is allowed to attend the unmasked
            positions. If a ByteTensor is provided, the non-zero positions are not allowed to attend
            while the zero positions will be unchanged. If a BoolTensor is provided, positions with ``True``
            are not allowed to attend while ``False`` values will be unchanged. If a FloatTensor
            is provided, it will be added to the attention weight.
            [src/tgt/memory]_key_padding_mask provides specified elements in the key to be ignored by
            the attention. If a ByteTensor is provided, the non-zero positions will be ignored while the zero
            positions will be unchanged. If a BoolTensor is provided, the positions with the
            value of ``True`` will be ignored while the position with the value of ``False`` will be unchanged.

            - output: :math:`(T, N, E)`.

            Note: Due to the multi-head attention architecture in the transformer model,
            the output sequence length of a transformer is same as the input sequence
            (i.e. target) length of the decode.

            where S is the source sequence length, T is the target sequence length, N is the
            batch size, E is the feature number

        Examples:
            >>> # output = transformer_model(src, tgt, src_mask=src_mask, tgt_mask=tgt_mask)
        """

        if src.shape[0] != tgt.shape[0]:
            raise RuntimeError("the batch number of src and tgt must be equal")

        # if src.shape[2] != self.head_size or tgt.shape[2] != self.head_size:
        #     raise RuntimeError(
        #         "the feature number of src and tgt must be equal to head_size"
        #     )

        if self.custom_encoder is not None:
            encoder = self.custom_encoder()
        else:
            encoder = TransformerEncoder(
                lambda: TransformerEncoderLayer(
                    self.head_size,
                    self.num_heads,
                    self.output_size,
                    self.dropout,
                    self.activation,
                ),
                self.num_encoder_layers,
                lambda: LayerNormalization(),
            )

        if self.custom_decoder is not None:
            decoder = self.custom_decoder()
        else:
            decoder = TransformerDecoder(
                lambda: TransformerDecoderLayer(
                    self.head_size,
                    self.num_heads,
                    self.output_size,
                    self.dropout,
                    self.activation,
                ),
                self.num_decoder_layers,
                lambda: LayerNormalization(),
            )

        memory = encoder(
            src,
            mask=src_mask,
            # src_key_padding_mask=src_key_padding_mask
        )
        output = decoder(
            tgt,
            memory,
            tgt_mask=tgt_mask,
            memory_mask=memory_mask,
            # tgt_key_padding_mask=tgt_key_padding_mask,
            # memory_key_padding_mask=memory_key_padding_mask,
        )

        return output



from hashlib import new
import unittest

import elegy
import jax
import jax.numpy as jnp
import numpy as np
import optax
import cloudpickle


class MLP(elegy.Module):
    """Standard LeNet-300-100 MLP network."""

    n1: int
    n2: int

    def __init__(self, n1: int = 3, n2: int = 4):
        super().__init__()
        self.n1 = n1
        self.n2 = n2

    def call(self, image: jnp.ndarray, training: bool):
        x = image.astype(jnp.float32) / 255.0

        x = jnp.reshape(x, [x.shape[0], -1])
        x = elegy.nn.Linear(self.n1)(x)
        x = elegy.nn.BatchNormalization()(x)
        x = jax.nn.relu(x)

        x = elegy.nn.Linear(self.n2)(x)
        x = jax.nn.relu(x)
        x = elegy.nn.Linear(10)(x)

        return x


class ModelBasicTest(unittest.TestCase):
    def test_predict(self):

        model = elegy.Model(module=elegy.nn.Linear(1))

        X = np.random.uniform(size=(5, 10))
        y = np.random.randint(10, size=(5, 1))

        y_pred = model.predict(x=X)

        assert y_pred.shape == (5, 1)

    def test_evaluate(self):
        def mse(y_true, y_pred):
            return jnp.mean((y_true - y_pred) ** 2)

        def mae(y_true, y_pred):
            return jnp.mean(jnp.abs(y_true - y_pred))

        model = elegy.Model(
            module=elegy.nn.Linear(1),
            loss=dict(a=mse),
            metrics=dict(b=mae),
            optimizer=optax.adamw(1e-3),
            run_eagerly=True,
        )

        X = np.random.uniform(size=(5, 10))
        y = np.random.uniform(size=(5, 1))

        logs = model.evaluate(x=X, y=y)

        assert "a/mse_loss" in logs
        assert "b/mae" in logs
        assert "loss" in logs

    def test_metrics(self):
        class M(elegy.Module):
            def call(self, x):
                n = self.add_parameter("n", lambda: 0, trainable=False)
                self.update_parameter("n", n + 1)
                return x

        metrics = elegy.model.model.Metrics(dict(a=dict(b=[M(), M()], c=M())))

        rng = elegy.RNGSeq(42)
        x = np.random.uniform(size=(5, 7, 7))

        with elegy.hooks.context(metrics=True):
            elegy.hooks.add_metric("d", 10)
            aux_metrics = elegy.hooks.get_metrics()
            logs, states = metrics.init(aux_metrics, rng)(x, training=True)

        with elegy.hooks.context(metrics=True):
            elegy.hooks.add_metric("d", 10)
            aux_metrics = elegy.hooks.get_metrics()
            logs, states = metrics.apply(aux_metrics, rng, states)(x, training=True)

        assert len(metrics.metrics) == 3
        assert "a/b/m" in metrics.metrics
        assert "a/b/m_1" in metrics.metrics
        assert "a/c/m" in metrics.metrics

        assert len(logs) == 4
        assert "a/b/m" in logs
        assert "a/b/m_1" in logs
        assert "a/c/m" in logs
        assert "d" in logs

        assert len(states) == 3
        assert "a/b/m" in states
        assert "a/b/m_1" in states
        assert "a/c/m" in states

    def test_losses(self):
        def loss_fn():
            return 3.0

        losses = elegy.model.model.Losses(dict(a=dict(b=[loss_fn, loss_fn], c=loss_fn)))

        rng = elegy.RNGSeq(42)
        hooks_losses = dict(x=0.3, y=4.5)

        with elegy.hooks.context(losses=True):
            elegy.hooks.add_loss("d", 1.0)
            aux_losses = elegy.hooks.get_losses()
            logs, logs, states = losses.init(aux_losses, rng)()

        with elegy.hooks.context(losses=True):
            elegy.hooks.add_loss("d", 1.0)
            aux_losses = elegy.hooks.get_losses()
            loss, logs, states = losses.apply(aux_losses, states)()

        assert loss == 10

        assert len(losses.losses) == 3
        assert "a/b/loss_fn" in losses.losses
        assert "a/b/loss_fn_1" in losses.losses
        assert "a/c/loss_fn" in losses.losses

        assert len(logs) == 5
        assert "loss" in logs
        assert "a/b/loss_fn" in logs
        assert "a/b/loss_fn_1" in logs
        assert "a/c/loss_fn" in logs
        assert "d_loss" in logs


class ModelTest(unittest.TestCase):
    def test_evaluate(self):

        model = elegy.Model(
            module=MLP(n1=3, n2=1),
            loss=[
                elegy.losses.SparseCategoricalCrossentropy(from_logits=True),
                elegy.regularizers.GlobalL2(l=1e-4),
            ],
            metrics=elegy.metrics.SparseCategoricalAccuracy(),
            optimizer=optax.adamw(1e-3),
            run_eagerly=True,
        )

        X = np.random.uniform(size=(5, 7, 7))
        y = np.random.randint(10, size=(5,))

        history = model.fit(
            x=X,
            y=y,
            epochs=1,
            steps_per_epoch=1,
            batch_size=5,
            validation_data=(X, y),
            shuffle=True,
            verbose=1,
        )

        logs = model.evaluate(X, y)

        eval_acc = logs["sparse_categorical_accuracy"]
        predict_acc = (model.predict(X).argmax(-1) == y).mean()

        assert eval_acc == predict_acc

    def test_cloudpickle(self):
        model = elegy.Model(
            module=MLP(n1=3, n2=1),
            loss=[
                elegy.losses.SparseCategoricalCrossentropy(from_logits=True),
                elegy.regularizers.GlobalL2(l=1e-4),
            ],
            metrics=elegy.metrics.SparseCategoricalAccuracy(),
            optimizer=optax.adamw(1e-3),
            run_eagerly=True,
        )

        X = np.random.uniform(size=(5, 7, 7))

        y0 = model.predict(X)

        model_pkl = cloudpickle.dumps(model)
        newmodel = cloudpickle.loads(model_pkl)

        newmodel.states = model.states
        newmodel.initial_states = model.initial_states

        y1 = newmodel.predict(X)
        assert np.all(y0 == y1)



from .data_handler import DataHandler
from .utils import (
    unpack_x_y_sample_weight,
    train_validation_split,
    map_structure,
    map_append,
)

from .dataset import Dataset, DataLoader


__all__ = ["Dataset", "DataLoader"]



from .binary_crossentropy import BinaryCrossentropy, binary_crossentropy
from .categorical_crossentropy import CategoricalCrossentropy
from .cosine_similarity import CosineSimilarity, cosine_similarity
from .huber import Huber, huber
from .loss import Loss, Reduction
from .mean_absolute_error import MeanAbsoluteError, mean_absolute_error
from .mean_absolute_percentage_error import (
    MeanAbsolutePercentageError,
    mean_absolute_percentage_error,
)
from .mean_squared_error import MeanSquaredError, mean_squared_error
from .mean_squared_logarithmic_error import (
    MeanSquaredLogarithmicError,
    mean_squared_logarithmic_error,
)
from .sparse_categorical_crossentropy import (
    SparseCategoricalCrossentropy,
    sparse_categorical_crossentropy,
)

__all__ = [
    "BinaryCrossentropy",
    "binary_crossentropy",
    "CategoricalCrossentropy",
    "CosineSimilarity",
    "cosine_similarity",
    "Huber",
    "huber",
    "Loss",
    "Reduction",
    "MeanAbsoluteError",
    "mean_absolute_error",
    "MeanAbsolutePercentageError",
    "mean_absolute_percentage_error",
    "MeanSquaredError",
    "mean_squared_error",
    "MeanSquaredLogarithmicError",
    "mean_squared_logarithmic_error",
    "SparseCategoricalCrossentropy",
    "sparse_categorical_crossentropy",
]



# Implementation based on tf.keras.callbacks.py
# https://github.com/tensorflow/tensorflow/blob/v2.2.0/tensorflow/python/keras/callbacks.py
import typing as tp

import numpy as np

from .callback import Callback


class LambdaCallback(Callback):
    r"""Callback for creating simple, custom callbacks on-the-fly.

    This callback is constructed with anonymous functions that will be called
    at the appropriate time. Note that the callbacks expects positional
    arguments, as:

    - `on_epoch_begin` and `on_epoch_end` expect two positional arguments:
        `epoch`, `logs`
    - `on_train_batch_begin` and `on_train_batch_end` expect two positional arguments:
        `batch`, `logs`
    - `on_train_begin` and `on_train_end` expect one positional argument:
        `logs`

    Example:

        ```python
        # Print the batch number at the beginning of every batch.
        batch_print_callback = LambdaCallback(
            on_train_batch_begin=lambda batch,logs: print(batch))

        # Stream the epoch loss to a file in JSON format. The file content
        # is not well-formed JSON but rather has a JSON object per line.
        import json
        json_log = open('loss_log.json', mode='wt', buffering=1)
        json_logging_callback = LambdaCallback(
            on_epoch_end=lambda epoch, logs: json_log.write(
                json.dumps({'epoch': epoch, 'loss': logs['loss']}) + '\n'),
            on_train_end=lambda logs: json_log.close()
        )

        # Terminate some processes after having finished model training.
        processes = ...
        cleanup_callback = LambdaCallback(
            on_train_end=lambda logs: [
                p.terminate() for p in processes if p.is_alive()])

        model.fit(...,
                    callbacks=[batch_print_callback,
                            json_logging_callback,
                            cleanup_callback])
        ```
    """

    def __init__(
        self,
        on_epoch_begin: tp.Optional[
            tp.Callable[[int, tp.Dict[str, np.ndarray]], None]
        ] = None,
        on_epoch_end: tp.Optional[
            tp.Callable[[int, tp.Dict[str, np.ndarray]], None]
        ] = None,
        on_train_batch_begin: tp.Optional[
            tp.Callable[[int, tp.Dict[str, np.ndarray]], None]
        ] = None,
        on_train_batch_end: tp.Optional[
            tp.Callable[[int, tp.Dict[str, np.ndarray]], None]
        ] = None,
        on_train_begin: tp.Optional[
            tp.Callable[[tp.Dict[str, np.ndarray]], None]
        ] = None,
        on_train_end: tp.Optional[tp.Callable[[tp.Dict[str, np.ndarray]], None]] = None,
        **kwargs
    ):
        """
        Arguments:
            on_epoch_begin: called at the beginning of every epoch.
            on_epoch_end: called at the end of every epoch.
            on_train_batch_begin: called at the beginning of every batch.
            on_train_batch_end: called at the end of every batch.
            on_train_begin: called at the beginning of model training.
            on_train_end: called at the end of model training.
        """
        super(LambdaCallback, self).__init__()
        self.__dict__.update(kwargs)
        if on_epoch_begin is not None:
            self.on_epoch_begin = on_epoch_begin
        else:
            self.on_epoch_begin = lambda epoch, logs: None
        if on_epoch_end is not None:
            self.on_epoch_end = on_epoch_end
        else:
            self.on_epoch_end = lambda epoch, logs: None
        if on_train_batch_begin is not None:
            self.on_train_batch_begin = on_train_batch_begin
        else:
            self.on_train_batch_begin = lambda batch, logs: None
        if on_train_batch_end is not None:
            self.on_train_batch_end = on_train_batch_end
        else:
            self.on_train_batch_end = lambda batch, logs: None
        if on_train_begin is not None:
            self.on_train_begin = on_train_begin
        else:
            self.on_train_begin = lambda logs: None
        if on_train_end is not None:
            self.on_train_end = on_train_end
        else:
            self.on_train_end = lambda logs: None



from unittest import TestCase

import elegy
import jax.numpy as jnp
import pytest


class MetricTest(TestCase):
    def test_basic(self):
        class MAE(elegy.Metric):
            def call(self, y_true, y_pred):
                return jnp.abs(y_true - y_pred)

        y_true = jnp.array([1.0, 2.0, 3.0])
        y_pred = jnp.array([2.0, 3.0, 4.0])

        mae = MAE()

        loss = mae.call_with_defaults()(y_true, y_pred)

        assert jnp.alltrue(loss == jnp.array([1.0, 1.0, 1.0]))

    def test_slice(self):
        class MAE(elegy.Metric):
            def call(self, y_true, y_pred):
                return jnp.abs(y_true - y_pred)

        y_true = dict(a=jnp.array([1.0, 2.0, 3.0]))
        y_pred = dict(a=jnp.array([2.0, 3.0, 4.0]))

        mae = MAE(on="a")

        # raises because it doesn't use kwargs
        with pytest.raises(BaseException):
            sample_loss = mae.call_with_defaults()(y_true, y_pred)

        # raises because it doesn't use __call__ which filters
        with pytest.raises(BaseException):
            sample_loss = mae.call(y_true=y_true, y_pred=y_pred)

        loss = mae.call_with_defaults()(y_true=y_true, y_pred=y_pred)

        assert jnp.alltrue(loss == jnp.array([1.0, 1.0, 1.0]))



import pathlib
import pickle
import threading
import typing as tp
from contextlib import contextmanager
from copy import copy
from dataclasses import dataclass
from enum import Enum
from io import StringIO

import cloudpickle
import jax
import jax.numpy as jnp
import numpy as np
import optax
import yaml
from elegy import hooks, module, types, utils
from elegy.losses.loss import Loss
from elegy.metrics.metric import Metric
from tabulate import tabulate


class PredStep(tp.NamedTuple):
    y_pred: tp.Any
    states: types.States
    aux_losses: types.Logs
    aux_metrics: types.Logs
    summaries: types.Summaries

    @classmethod
    def simple(cls, y_pred: tp.Any, states: types.States):
        return cls(
            y_pred=y_pred,
            states=states,
            aux_losses=hooks.get_losses(),
            aux_metrics=hooks.get_metrics(),
            summaries=hooks.get_summaries(),
        )


class TestStep(tp.NamedTuple):
    loss: types.Scalar
    logs: types.Logs
    states: types.States


class GradStep(tp.NamedTuple):
    loss: types.Scalar
    logs: types.Logs
    states: types.States
    grads: types.Grads


class TrainStep(tp.NamedTuple):
    logs: types.Logs
    states: types.States


class ModelCore:
    states: types.States
    initial_states: types.States
    history: tp.Dict[str, tp.Any]
    run_eagerly: bool = False

    def __init__(
        self,
        net_params: tp.Union[types.Uninitialized, tp.Any] = types.UNINITIALIZED,
        net_states: tp.Union[types.Uninitialized, tp.Any] = types.UNINITIALIZED,
        metrics_states: tp.Union[types.Uninitialized, tp.Any] = types.UNINITIALIZED,
        optimizer_states: tp.Union[types.Uninitialized, tp.Any] = types.UNINITIALIZED,
        rng: tp.Union[types.Uninitialized, tp.Any] = types.UNINITIALIZED,
        run_eagerly: bool = False,
    ):
        self.states = types.States(
            net_params=net_params,
            net_states=net_states,
            metrics_states=metrics_states,
            optimizer_states=optimizer_states,
            rng=rng,
        )
        self.initial_states = self.states.copy()
        self.run_eagerly = run_eagerly
        self.history = {}

        self._jit_functions()

    def _jit_functions(self):
        self.call_pred_step_jit = hooks.jit(
            self.call_pred_step,
            static_argnums=[1, 3],
        )
        self.call_test_step_jit = hooks.jit(
            self.call_test_step,
            static_argnums=[2, 6],
        )
        self.call_train_step_jit = hooks.jit(
            self.call_train_step,
            static_argnums=[2, 6],
        )

    def __setstate__(self, d):
        self.__dict__ = d
        self._jit_functions()

    def __getstate__(self):
        d = self.__dict__.copy()

        # remove states
        del d["states"]
        del d["initial_states"]

        # remove jitted functions
        del d["call_pred_step_jit"]
        del d["call_test_step_jit"]
        del d["call_train_step_jit"]

        return d

    # ----------------------------------------------------------------
    # Abstract API
    # ----------------------------------------------------------------

    def pred_step(
        self,
        x: tp.Any,
        mode: types.Mode,
        net_params: tp.Any,
        net_states: tp.Any,
        rng: tp.Any,
        states: types.States,
        initializing: bool,
        training: bool,
    ) -> PredStep:
        raise NotImplementedError()

    def call_pred_step(
        self,
        x: tp.Any,
        mode: types.Mode,
        states: types.States,
        initializing: bool,
    ) -> PredStep:
        get_losses_and_metrics = mode in (types.Mode.test, types.Mode.train)
        get_summaries = mode == types.Mode.summary

        with hooks.context(
            losses=get_losses_and_metrics,
            metrics=get_losses_and_metrics,
            summaries=get_summaries,
        ):
            return utils.inject_dependencies(self.pred_step)(
                x=x,
                mode=mode,
                net_params=states.net_params,
                net_states=states.net_states,
                rng=states.rng,
                states=states,
                initializing=initializing,
                training=(mode == types.Mode.train),
            )

    def test_step(
        self,
        x: tp.Any,
        y_true: tp.Any,
        mode: types.Mode,
        net_params: tp.Any,
        net_states: tp.Any,
        metrics_states: tp.Any,
        sample_weight: tp.Optional[np.ndarray],
        class_weight: tp.Optional[np.ndarray],
        rng: tp.Any,
        states: types.States,
        initializing: bool,
        training: bool,
    ) -> TestStep:
        raise NotImplementedError()

    def call_test_step(
        self,
        x: tp.Any,
        y_true: tp.Any,
        mode: types.Mode,
        sample_weight: tp.Optional[np.ndarray],
        class_weight: tp.Optional[np.ndarray],
        states: types.States,
        initializing: bool,
    ) -> TestStep:
        return utils.inject_dependencies(self.test_step)(
            net_params=states.net_params,
            x=x,
            y_true=y_true,
            mode=mode,
            net_states=states.net_states,
            metrics_states=states.metrics_states,
            sample_weight=sample_weight,
            class_weight=class_weight,
            rng=states.rng,
            states=states,
            initializing=initializing,
            training=(mode == types.Mode.train),
        )

    def grad_step(
        self,
        x: tp.Any,
        y_true: tp.Any,
        mode: types.Mode,
        net_params: tp.Any,
        net_states: tp.Any,
        metrics_states: tp.Any,
        sample_weight: tp.Optional[np.ndarray],
        class_weight: tp.Optional[np.ndarray],
        rng: tp.Any,
        states: types.States,
        initializing: bool,
        training: bool,
    ) -> GradStep:
        raise NotImplementedError()

    def call_grad_step(
        self,
        x: tp.Any,
        y_true: tp.Any,
        mode: types.Mode,
        sample_weight: tp.Optional[np.ndarray],
        class_weight: tp.Optional[np.ndarray],
        states: types.States,
        initializing: bool,
    ) -> GradStep:
        return utils.inject_dependencies(self.grad_step)(
            x=x,
            y_true=y_true,
            mode=mode,
            net_params=states.net_params,
            net_states=states.net_states,
            metrics_states=states.metrics_states,
            sample_weight=sample_weight,
            class_weight=class_weight,
            rng=states.rng,
            states=states,
            initializing=initializing,
            training=(mode == types.Mode.train),
        )

    def train_step(
        self,
        x: tp.Any,
        y_true: tp.Any,
        mode: types.Mode,
        net_params: tp.Any,
        net_states: tp.Any,
        metrics_states: tp.Any,
        optimizer_states: tp.Any,
        sample_weight: tp.Optional[np.ndarray],
        class_weight: tp.Optional[np.ndarray],
        rng: tp.Any,
        states: types.States,
        initializing: bool,
        training: bool,
    ) -> TrainStep:
        raise NotImplementedError()

    def call_train_step(
        self,
        x: tp.Any,
        y_true: tp.Any,
        mode: types.Mode,
        sample_weight: tp.Optional[np.ndarray],
        class_weight: tp.Optional[np.ndarray],
        states: types.States,
        initializing: bool,
    ) -> TrainStep:
        return utils.inject_dependencies(self.train_step)(
            x=x,
            y_true=y_true,
            mode=mode,
            net_params=states.net_params,
            net_states=states.net_states,
            metrics_states=states.metrics_states,
            optimizer_states=states.optimizer_states,
            sample_weight=sample_weight,
            class_weight=class_weight,
            rng=states.rng,
            states=states,
            initializing=initializing,
            training=(mode == types.Mode.train),
        )

    # ----------------------------------------------------------------
    # *_on_batch methods
    # ----------------------------------------------------------------

    def predict_on_batch(
        self, x: tp.Union[np.ndarray, tp.Mapping[str, tp.Any], tp.Tuple]
    ) -> tp.Union[np.ndarray, tp.Mapping[str, tp.Any], tp.Tuple]:
        """
        Returns predictions for a single batch of samples.

        Arguments:
            x: Input data. A Numpy/Jax array (or array-like), or possibly
                nested python structure of dict, list, tuple that contain
                arrays as leafs.

        Returns:
            Jax array(s) of predictions.

        Raises:
            ValueError: In case of mismatch between given number of inputs and
                expectations of the model.
        """
        mode = types.Mode.pred
        initializing = False

        self.maybe_initialize(mode=mode, x=x)

        method = self.call_pred_step if self.run_eagerly else self.call_pred_step_jit

        y_pred, state_updates, _, _, _ = method(x, mode, self.states, initializing)

        self.states = self.states.merge(state_updates)

        return y_pred

    def test_on_batch(
        self,
        x: tp.Union[np.ndarray, tp.Mapping[str, tp.Any], tp.Tuple],
        y: tp.Union[np.ndarray, tp.Mapping[str, tp.Any], tp.Tuple, None] = None,
        sample_weight: tp.Optional[np.ndarray] = None,
        class_weight: tp.Optional[np.ndarray] = None,
    ) -> types.Logs:
        """
        Test the model on a single batch of samples.

        Arguments:
            x: Input data. It could be:

                - A Numpy array (or array-like), or a list
                    of arrays (in case the model has multiple inputs).
                - A dict mapping input names to the corresponding arrays, if
                    the model has named inputs.
            y: Target data. Like the input data `x`, it could be either Numpy
                array(s) or Jax array(s).
            sample_weight: Optional array of the same length as x, containing
                weights to apply to the model's loss for each sample. In the case of
                temporal data, you can pass a 2D array with shape (samples,
                sequence_length), to apply a different weight to every timestep of
                every sample.

        Returns:
            A `logs` dictionary of containing the main `loss` as well as all
            other losses and metrics.
        Raises:
            ValueError: In case of invalid user-provided arguments.
        """
        mode = types.Mode.test
        initializing = False

        self.maybe_initialize(
            mode,
            x=x,
            y_true=y,
            sample_weight=sample_weight,
            class_weight=class_weight,
        )

        method = self.call_test_step if self.run_eagerly else self.call_test_step_jit

        loss, logs, state_updates = method(
            x,
            y,
            mode,
            sample_weight,
            class_weight,
            self.states,
            initializing,
        )

        self.states = self.states.merge(state_updates)

        return logs

    def train_on_batch(
        self,
        x: tp.Union[np.ndarray, tp.Mapping[str, tp.Any], tp.Tuple],
        y: tp.Union[np.ndarray, tp.Mapping[str, tp.Any], tp.Tuple, None] = None,
        sample_weight: tp.Optional[np.ndarray] = None,
        class_weight: tp.Optional[np.ndarray] = None,
    ) -> types.Logs:
        """
        Runs a single gradient update on a single batch of data.

        Arguments:
            x: Input data. It could be:

                - A Numpy array (or array-like), or a iterable of arrays
                    (in case the model has multiple inputs).
                - A dict mapping input names to the corresponding arrays,
                    if the model has named inputs.
            y: Target data. Like the input data `x`, it could be either Numpy
                array(s) or Jax array(s). It should be consistent with `x`
                (you cannot have Numpy inputs and array targets, or inversely).
            sample_weight: Optional array of the same length as x, containing
                weights to apply to the model's loss for each sample. In the case of
                temporal data, you can pass a 2D array with shape (samples,
                sequence_length), to apply a different weight to every timestep of
                every sample.
            class_weight: Optional dictionary mapping class indices (integers) to a
                weight (float) to apply to the model's loss for the samples from this
                class during training. This can be useful to tell the model to "pay
                more attention" to samples from an under-represented class.

        Returns:
            A `logs` dictionary of containing the main `loss` as well as all
            other losses and metrics.

        Raises:
            ValueError: In case of invalid user-provided arguments.
        """
        mode = types.Mode.train
        initializing = False

        self.maybe_initialize(
            mode=mode,
            x=x,
            y_true=y,
            sample_weight=sample_weight,
            class_weight=class_weight,
        )

        method = self.call_train_step if self.run_eagerly else self.call_train_step_jit

        logs, state_updates = method(
            x,
            y,
            mode,
            sample_weight,
            class_weight,
            self.states,
            initializing,
        )

        self.states = self.states.merge(state_updates)

        return logs

    # ----------------------------------------------------------------
    # other methods
    # ----------------------------------------------------------------

    def update_modules(self):
        pass

    def reset(self):
        self.states = self.initial_states.copy()

    def reset_metrics(self):
        self.states = self.states.update(
            metrics_states=self.initial_states.metrics_states
        )

    def maybe_initialize(
        self,
        mode: types.Mode,
        x: tp.Union[np.ndarray, tp.Mapping[str, tp.Any], tp.Tuple] = (),
        y_true: tp.Union[np.ndarray, tp.Mapping[str, tp.Any], tp.Tuple, None] = None,
        sample_weight: tp.Optional[np.ndarray] = None,
        class_weight: tp.Optional[np.ndarray] = None,
    ):
        rng = self.states.rng if isinstance(self.states.rng, types.RNGSeq) else None
        initializing = True
        state_updates: types.States

        if (
            mode in (types.Mode.pred, types.Mode.summary)
            and isinstance(self.states.net_params, types.Uninitialized)
            and isinstance(self.states.net_states, types.Uninitialized)
        ):
            method = (
                self.call_pred_step if self.run_eagerly else self.call_pred_step_jit
            )

            _, state_updates, _, _, _ = method(
                x,
                mode,
                self.states,
                initializing,
            )
        elif mode == types.Mode.test and isinstance(
            self.states.metrics_states, types.Uninitialized
        ):
            method = (
                self.call_test_step if self.run_eagerly else self.call_test_step_jit
            )

            _, _, state_updates = method(
                x,
                y_true,
                mode,
                sample_weight,
                class_weight,
                self.states,
                initializing,
            )
        elif mode == types.Mode.train and isinstance(
            self.states.optimizer_states, types.Uninitialized
        ):
            method = (
                self.call_train_step if self.run_eagerly else self.call_train_step_jit
            )

            _, state_updates = method(
                x,
                y_true,
                mode,
                sample_weight,
                class_weight,
                self.states,
                initializing,
            )

        else:
            return

        if mode in (types.Mode.pred, types.Mode.test, types.Mode.train):
            if isinstance(state_updates.net_params, types.Uninitialized):
                state_updates = state_updates.update(net_params=None)

            if isinstance(state_updates.net_states, types.Uninitialized):
                state_updates = state_updates.update(net_states=None)
        if mode in (types.Mode.test, types.Mode.train):
            if isinstance(state_updates.metrics_states, types.Uninitialized):
                state_updates = state_updates.update(metrics_states=None)

        if mode == types.Mode.train:
            if isinstance(state_updates.optimizer_states, types.Uninitialized):
                state_updates = state_updates.update(optimizer_states=None)

        self.states = self.states.merge_new(state_updates)
        self.initial_states = self.initial_states.merge_new(state_updates)

    def save(
        self,
        path: tp.Union[str, pathlib.Path],
    ) -> None:
        """
        Saves the model to disk.

        It creates a directory that includes:

        - `{path}/model.pkl`: The `Model` object instance serialized with `pickle`,
            this allows you to re-instantiate the model later.
        - `{path}/states.pkl`: The `Model.states` serialized with `pickle`.
        - `{path}/initial_states.pkl`: The `Model.initial_states` serialized with `pickle`.

        This allows you to save the entirety of the states of a model
        in a directory structure which can be fully restored via
        `Model.load` if the model is already instiated or `elegy.model.load`
        to load the model instance from its pickled version.

        ```python
        import elegy

        model.save('my_model')  # creates folder at 'my_model'
        del model  # deletes the existing model

        # returns a model identical to the previous one
        model = elegy.model.load('my_model')
        ```
        Arguments:
            path: path where model structure will be saved.
        """
        if isinstance(path, str):
            path = pathlib.Path(path)

        path.mkdir(parents=True, exist_ok=True)

        with open(path / "states.pkl", "wb") as f:
            cloudpickle.dump(self.states, f)

        with open(path / "initial_states.pkl", "wb") as f:
            cloudpickle.dump(self.initial_states, f)

        path = path / "model.pkl"

        with open(path, "wb") as f:
            try:
                cloudpickle.dump(self, f)
            except BaseException as e:
                print(
                    f"Error occurred saving the model object at {path}\nContinuing...."
                )

    def load(
        self,
        path: tp.Union[str, pathlib.Path],
    ) -> None:
        """
        Loads all weights + states from a folder structure.

        You can load states from other models that have slightly different architecture
        as long as long as it preserves the ordering of the `haiku.Params` + `haiku.State`
        structures, adding or removing layers is fine as long as they don't have weights,
        new layers with weights will be initialized from scratch.

        Arguments:
            path: path to a saved model's directory.
        """
        if isinstance(path, str):
            path = pathlib.Path(path)

        self.states = cloudpickle.loads((path / "states.pkl").read_bytes())
        self.initial_states = cloudpickle.loads(
            (path / "initial_states.pkl").read_bytes()
        )



import typing as tp

import jax
import numpy as np
from jax import numpy as jnp

from elegy import initializers, module
from elegy.nn.dropout import Dropout
from elegy.nn.layer_normalization import LayerNormalization
from elegy.nn.linear import Linear
from elegy.nn.sequential_module import sequential
from elegy import types


class MultiHeadAttention(module.Module):
    r"""
    MultiHead Attention layer.
    Defines the MultiHead Attention operation as described in
    [Attention Is All You Need](https://arxiv.org/abs/1706.03762) which takes
    in the tensors `query`, `key`, and `value`, and returns the dot-product attention
    between them:

    ```python
    mha = MultiHeadAttention(head_size=128, num_heads=12)
    query = tf.random.uniform((32, 20, 200)) # (batch_size, query_elements, query_depth)
    key = tf.random.uniform((32, 15, 300)) # (batch_size, key_elements, key_depth)
    value = tf.random.uniform((32, 15, 400)) # (batch_size, key_elements, value_depth)
    attention = mha([query, key, value]) # (batch_size, query_elements, value_depth)
    ```

    If `value` is not given then internally `value = key` will be used:

    ```python
    mha = MultiHeadAttention(head_size=128, num_heads=12)
    query = tf.random.uniform((32, 20, 200)) # (batch_size, query_elements, query_depth)
    key = tf.random.uniform((32, 15, 300)) # (batch_size, key_elements, key_depth)
    attention = mha([query, key]) # (batch_size, query_elements, key_depth)
    ```

    Arguments:
        head_size: int, dimensionality of the `query`, `key` and `value` tensors
        after the linear transformation.
        num_heads: int, number of attention heads.
        output_size: int, dimensionality of the output space, if `None` then the
        input dimension of
        `value` or `key` will be used, default `None`.
        dropout: float, `rate` parameter for the dropout layer that is
        applied to attention after softmax,
        default `0`.
        use_projection_bias: bool, whether to use a bias term after the linear
        output projection.
        return_attn_coef: bool, if `True`, return the attention coefficients as
        an additional output argument.
        kernel_initializer: initializer, initializer for the kernel weights.
        kernel_regularizer: regularizer, regularizer for the kernel weights.
        kernel_constraint: constraint, constraint for the kernel weights.
        bias_initializer: initializer, initializer for the bias weights.
        bias_regularizer: regularizer, regularizer for the bias weights.
        bias_constraint: constraint, constraint for the bias weights.

    """

    def __init__(
        self,
        head_size: int,
        num_heads: int,
        output_size: tp.Optional[int] = None,
        dropout: float = 0.0,
        use_projection_bias: bool = True,
        return_attn_coef: bool = False,
        kernel_initializer: types.Initializer = initializers.VarianceScaling(scale=2.0),
        bias_initializer: types.Initializer = initializers.Constant(0.0),
        # kernel_initializer: typing.Union[str, typing.Callable] = "glorot_uniform",
        # kernel_regularizer: typing.Union[str, typing.Callable] = None,
        # kernel_constraint: typing.Union[str, typing.Callable] = None,
        # bias_regularizer: typing.Union[str, typing.Callable] = None,
        # bias_constraint: typing.Union[str, typing.Callable] = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        if output_size is not None and output_size < 1:
            raise ValueError("output_size must be a positive number")

        self.head_size = head_size
        self.num_heads = num_heads
        self.output_size = output_size
        self.use_projection_bias = use_projection_bias
        self.return_attn_coef = return_attn_coef
        self.droput_rate = dropout

        self.kernel_initializer = kernel_initializer
        self.bias_initializer = bias_initializer

    def call(
        self,
        query: jnp.ndarray,
        key: tp.Optional[jnp.ndarray] = None,
        value: tp.Optional[jnp.ndarray] = None,
        mask=None,
        training=None,
    ):
        """
        Arguments:
            inputs:  List of `[query, key, value]` where
                * `query`: np.ndarray of shape `(..., query_elements, query_depth)`
                * `key`: `np.ndarray of shape '(..., key_elements, key_depth)`
                * `value`: np.ndarray of shape `(..., key_elements, value_depth)`, optional, if not given `key` will be used.
            mask: a binary np.ndarray of shape `[batch_size?, num_heads?, query_elements, key_elements]`
                which specifies which query elements can attendo to which key elements,
                `1` indicates attention and `0` indicates no attention.
        Output shape:
            * `(..., query_elements, output_size)` if `output_size` is given, else
            * `(..., query_elements, value_depth)` if `value` is given, else
            * `(..., query_elements, key_depth)`
        """

        # einsum nomenclature
        # ------------------------
        # N = query elements
        # M = key/value elements
        # H = heads
        # I = input features
        # O = output features

        if key is None:
            key = query

        if value is None:
            value = key

        output_size = (
            self.output_size if self.output_size is not None else value.shape[-1]
        )

        # verify shapes
        if key.shape[-2] != value.shape[-2]:
            raise ValueError(
                "the number of elements in 'key' must be equal to the same as the number of elements in 'value'"
            )

        if mask is not None:
            if len(mask.shape) < 2:
                raise ValueError("'mask' must have atleast 2 dimensions")
            if query.shape[-2] != mask.shape[-2]:
                raise ValueError(
                    "mask's second to last dimension must be equal to the number of elements in 'query'"
                )
            if key.shape[-2] != mask.shape[-1]:
                raise ValueError(
                    "mask's last dimension must be equal to the number of elements in 'key'"
                )

        # get weights
        query_kernel = self.add_parameter(
            "query_kernel",
            [self.num_heads, query.shape[-1], self.head_size],
            jnp.float32,
            initializer=self.kernel_initializer,
        )
        key_kernel = self.add_parameter(
            "key_kernel",
            [self.num_heads, key.shape[-1], self.head_size],
            jnp.float32,
            initializer=self.kernel_initializer,
        )
        value_kernel = self.add_parameter(
            "value_kernel",
            [self.num_heads, value.shape[-1], self.head_size],
            jnp.float32,
            initializer=self.kernel_initializer,
        )
        projection_kernel = self.add_parameter(
            "projection_kernel",
            [self.num_heads, self.head_size, output_size],
            jnp.float32,
            initializer=self.kernel_initializer,
        )

        # Linear transformations
        query = jnp.einsum("...NI , HIO -> ...NHO", query, query_kernel)
        key = jnp.einsum("...MI , HIO -> ...MHO", key, key_kernel)
        value = jnp.einsum("...MI , HIO -> ...MHO", value, value_kernel)

        # Scale dot-product, doing the division to either query or key
        # instead of their product saves some computation
        query /= jnp.sqrt(self.head_size)

        # Calculate dot product attention
        logits = jnp.einsum("...NHO,...MHO->...HNM", query, key)

        # apply mask
        if mask is not None:
            mask = mask.astype(jnp.float32)

            # possibly expand on the head dimension so broadcasting works
            if len(mask.shape) != len(logits.shape):
                mask = jnp.expand_dims(mask, -3)

            logits += -10e9 * (1.0 - mask)

        attn_coef = jax.nn.softmax(logits)

        # attention dropout
        attn_coef_dropout = Dropout(self.droput_rate)(attn_coef, training=training)

        # attention * value
        multihead_output = jnp.einsum("...HNM,...MHI->...NHI", attn_coef_dropout, value)

        # Run the outputs through another linear projection layer. Recombining heads
        # is automatically done.
        output = jnp.einsum("...NHI,HIO->...NO", multihead_output, projection_kernel)

        if self.use_projection_bias:
            output += self.add_parameter(
                "projection_bias",
                [output_size],
                jnp.float32,
                initializer=self.bias_initializer,
            )

        if self.return_attn_coef:
            return output, attn_coef
        else:
            return output



import jax.numpy as jnp

from unittest import TestCase

import elegy


class BatchNormalizationTest(TestCase):
    def test_connects(self):
        elegy.nn.BatchNormalization().call_with_defaults()(
            jnp.ones([3, 3]),
            training=True,
        )

    def test_on_predict(self):
        class TestModule(elegy.Module):
            def call(self, x, training):
                return elegy.nn.BatchNormalization()(x, training)

        model = elegy.Model(module=TestModule())

        x = jnp.ones([3, 5])

        y_pred = model.predict(x)
        logs = model.evaluate(x)



from unittest import TestCase

import numpy as np

import elegy


class LinearTest(TestCase):
    def test_connects(self):

        x = np.random.uniform(-1, 1, size=(4, 3))
        linear = elegy.nn.Linear(5)

        y_pred = linear.call_with_defaults(rng=elegy.RNGSeq(42))(x)

        assert y_pred.shape == (4, 5)

    def test_on_model(self):

        model = elegy.Model(module=elegy.nn.Linear(2))

        x = np.ones([3, 5])

        y_pred = model.predict(x)
        logs = model.evaluate(x)



import jax
import elegy
import unittest
import numpy as np
import jax.numpy as jnp

import optax


class MLP(elegy.Module):
    """Standard LeNet-300-100 MLP network."""

    n1: int
    n2: int

    def __init__(self, n1: int = 3, n2: int = 4):
        super().__init__()
        self.n1 = n1
        self.n2 = n2

    def call(self, image: jnp.ndarray, training: bool):
        x = image.astype(jnp.float32) / 255.0

        x = jnp.reshape(x, [x.shape[0], -1])
        x = elegy.nn.Linear(self.n1)(x)
        x = elegy.nn.BatchNormalization()(x)
        x = jax.nn.relu(x)

        x = elegy.nn.Linear(self.n2)(x)
        x = jax.nn.relu(x)
        x = elegy.nn.Linear(10)(x)

        return x


class OptimizerTest(unittest.TestCase):
    def test_optimizer(self):
        optax_op = optax.adam(1e-3)
        lr_schedule = lambda step, epoch: step / 3

        optimizer = elegy.Optimizer(optax_op, lr_schedule=lr_schedule)

        params = np.random.uniform((3, 4))
        grads = np.random.uniform((3, 4))
        rng = elegy.RNGSeq(42)

        optimizer_states = optimizer.init(rng, params)
        assert jnp.allclose(optimizer.current_lr(optimizer_states), 0 / 3)

        params, optimizer_states = optimizer.apply(params, grads, optimizer_states, rng)
        assert jnp.allclose(optimizer.current_lr(optimizer_states), 1 / 3)

        params, optimizer_states = optimizer.apply(params, grads, optimizer_states, rng)
        assert jnp.allclose(optimizer.current_lr(optimizer_states), 2 / 3)

        params, optimizer_states = optimizer.apply(params, grads, optimizer_states, rng)
        assert jnp.allclose(optimizer.current_lr(optimizer_states), 3 / 3)

    def test_optimizer_epoch(self):
        optax_op = optax.adam(1e-3)
        lr_schedule = lambda step, epoch: epoch

        optimizer = elegy.Optimizer(
            optax_op, lr_schedule=lr_schedule, steps_per_epoch=2
        )

        params = np.random.uniform((3, 4))
        grads = np.random.uniform((3, 4))
        rng = elegy.RNGSeq(42)

        optimizer_states = optimizer.init(
            rng=rng,
            net_params=params,
        )

        assert jnp.allclose(optimizer.current_lr(optimizer_states), 0)
        params, optimizer_states = optimizer.apply(params, grads, optimizer_states, rng)

        assert jnp.allclose(optimizer.current_lr(optimizer_states), 0)
        params, optimizer_states = optimizer.apply(params, grads, optimizer_states, rng)

        assert jnp.allclose(optimizer.current_lr(optimizer_states), 1)
        params, optimizer_states = optimizer.apply(params, grads, optimizer_states, rng)

        assert jnp.allclose(optimizer.current_lr(optimizer_states), 1)
        params, optimizer_states = optimizer.apply(params, grads, optimizer_states, rng)

    def test_optimizer_chain(self):

        optimizer = elegy.Optimizer(
            optax.sgd(0.1),
            optax.clip(0.5),
        )

        params = np.zeros(shape=(3, 4))
        grads = np.ones(shape=(3, 4)) * 100_000
        rng = elegy.RNGSeq(42)

        optimizer_states = optimizer.init(
            rng=rng,
            net_params=params,
        )

        params, optimizer_states = optimizer.apply(params, grads, optimizer_states, rng)

        assert np.all(-0.5 <= params) and np.all(params <= 0.5)

    def test_lr_logging(self):
        model = elegy.Model(
            module=MLP(n1=3, n2=1),
            loss=elegy.losses.SparseCategoricalCrossentropy(from_logits=True),
            metrics=elegy.metrics.SparseCategoricalAccuracy(),
            optimizer=elegy.Optimizer(
                optax.adamw(1.0, b1=0.95),
                lr_schedule=lambda step, epoch: jnp.array(1e-3),
            ),
            run_eagerly=True,
        )

        X = np.random.uniform(size=(5, 7, 7))
        y = np.random.randint(10, size=(5,))

        history = model.fit(
            x=X,
            y=y,
            epochs=1,
            steps_per_epoch=1,
            batch_size=5,
            validation_data=(X, y),
            shuffle=True,
            verbose=0,
        )

        assert "lr" in history.history
        assert np.allclose(history.history["lr"], 1e-3)
